// Draconian ARM routines
// Last Update: September 24, 2017
//              Darrell Spice, Jr.


// uncomment the following if using getPitch.  It must be before defines_cdf.h
//#define GETPITCH

#include "defines.h"
#include "defines_dasm.h"
//#include "defines_cdf_v1.h"
#include "defines_cdf_v1_1.h"
#include "alpha_levels.h"
#include "beta_levels.h"
#include "gamma_levels.h"
#include "delta_levels.h"

// work is spread across frames due to performance reasons
//
// FRAME & 3 FUNCTION           FEATURE - work spread across 4 frames
// --------- ------------------ -------------------------------------
//     0     StationMissiles()  fire new shots from stations 1, 5
//           MoveEnemyShips()   spawn I-Type missile
//     1     StationMissiles()  fire new shots from stations 2, 6
//           MoveEnemyShips()   spawn P-Type missile
//     2     StationMissiles()  fire new shots from stations 3, 7
//           MoveFormation()    spawn Formation
//     3     StationMissiles()  fire new shots from stations 4, 8
//           MoveSpyShip()      spawn Spy Ship
//
// FRAME & 7 FUNCTION           FEATURE - work spread across 8 frames
// --------- ------------------ -------------------------------------
//     0     StationOpenClose() Open/Close core of Station 1
//           MoveEnemyShips()   move I&P-Type 24
//           MoveFormation()    update direction of formation ships
//     1     StationOpenClose() Open/Close core of Station 2
//           MoveEnemyShips()   move I&P-Type 25
//           MoveSpyShip()      update direction of spy ship
//     2     StationOpenClose() Open/Close core of Station 3
//           MoveEnemyShips()   move I&P-Type 26
//     3     StationOpenClose() Open/Close core of Station 4
//           MoveEnemyShips()   move I&P-Type 27
//     4     StationOpenClose() Open/Close core of Station 5
//           MoveEnemyShips()   move I&P-Type 28
//     5     StationOpenClose() Open/Close core of Station 6
//           MoveEnemyShips()   move I&P-Type 29
//     6     StationOpenClose() Open/Close core of Station 7
//           MoveEnemyShips()   move I&P-Type 30
//     7     StationOpenClose() Open/Close core of Station 8
//           MoveEnemyShips()   move I&P-Type 23



#ifdef SHOW_AUDIO_PADDING
unsigned int gPadding;
#endif



unsigned int gScore;
unsigned int gLastScore;
unsigned int gHighScore;
unsigned int gExtraLife;

#ifdef SAVE_HIGH_SCORES
unsigned int gHighScores[4][4];
#endif

unsigned int gSector;
unsigned int gSectorBCD;
unsigned char * gMenuDigitGfx = (unsigned char *)MENU_DIGITS;
unsigned char gCurrentSample;     //unsigned int gCurrentSample; used 8 more bytes
//unsigned char gCondition;
unsigned char gReserveLives;
unsigned short int gActiveSound[2];
unsigned char gSoundDuration[2];
unsigned short int gPriorFrameStationOnscreen;
unsigned char gPriorFrameEnemyMissileOnscreen;

unsigned char gRoundStart;        // non-zero means round is starting up
                                  // 0 = round fully active
                                  // 1 = wait for end of opening tune then Say "Blast Off" (show Score/Radar/Lives)
                                  // 2 = wait for end of "Blast Off" then start ship moving

unsigned int gUnusedExplosionSprite;  // saved 12 bytes ROM vs char
unsigned int gUnusedEnemySprite;      // saved 12 bytes ROM vs char
unsigned char gPaused;
unsigned char gColorSwitchState;
unsigned char gSystemIs7800;
unsigned char gAlertRepeat;
unsigned short int gGameDelay;

// set if player is holding fire at end of game so next game doesn't start right away
unsigned char gIgnoreFireInMenu;


unsigned short int gEnemySpeed;
unsigned short int gEnemyTimer[2];    // countdown timer for launcing of P and I type ships
unsigned char gEnemyMax[2];           // how many P and I type ships
unsigned short int gEnemyMinTime;
unsigned short int gEnemyRangeTime;

unsigned short int gFormationTimer;     // countdown timer for launcing of formations
unsigned short int gFormationMinTime;
unsigned short int gFormationRangeTime;
unsigned char gFormationShipType;
unsigned short int gFormationSpeed;
unsigned char gFormationMask;           // used to select a random formation

unsigned short int gSpyTimer;    // countdown timer for launcing of spy ships
unsigned short int gSpyMinTime;
unsigned short int gSpyRangeTime;
unsigned short int gSpySpeed;
unsigned char gSpyPasses;
unsigned char gSpyBonus;

unsigned short int gFormationSpyCountdown; // countdown of formatins remaining until a spy ship can appear

unsigned short int gNotSoChatty;  // reduce frequency of saying Alert! ALert!

unsigned short int gIdleConditionRedMax;
unsigned short int gIdleConditionRedTimer; // triggers condition red if player's just flying around ignoring the stations

// location of Sector that's currently onscreen
unsigned short int gWindowX;
unsigned short int gWindowY;

// Game Screen Data Streams
unsigned char * gP0DataStream             =(unsigned char *)(DD_BASE + P0_DATASTREAM);
unsigned char * gP1DataStream             =(unsigned char *)(DD_BASE + P1_DATASTREAM);
unsigned char * gHMP0DataStream           =(unsigned char *)(DD_BASE + HMP0_DATASTREAM);
unsigned char * gHMP1DataStream           =(unsigned char *)(DD_BASE + HMP1_DATASTREAM);
unsigned char * gMissile0DataStream       =(unsigned char *)(DD_BASE + MISSILE0_DATASTREAM);
unsigned char * gMissile1DataStream       =(unsigned char *)(DD_BASE + MISSILE1_DATASTREAM);
unsigned char * gBallDataStream           =(unsigned char *)(DD_BASE + BALL_DATASTREAM);
unsigned char * gColorDataStream          =(unsigned char *)(DD_BASE + COLOR_DATASTREAM);
unsigned char * gSizeDataStream           =(unsigned char *)(DD_BASE + SIZE_DATASTREAM);
unsigned short int * gJumpDataStream      =(unsigned short int *)(DD_BASE + JUMP_DATASTREAM);
unsigned short int * gScoreDataStream     =(unsigned short int *)(DD_BASE + SCORE_DATASTREAM);

// Score/Radar display data
unsigned char * gPlayfield        =(unsigned char *)(DD_BASE + PLAYFIELD_DATA);
unsigned char * gRadarControl     =(unsigned char *)(DD_BASE + RADAR_CONTROL);
unsigned char * gRadarP0          =(unsigned char *)(DD_BASE + RADAR_P0);
unsigned char * gRadarP1          =(unsigned char *)(DD_BASE + RADAR_P1);
unsigned char * gFormationP0      =(unsigned char *)(DD_BASE + FORMATION_P0);
//unsigned char * gFormationP1      =(unsigned char *)(DD_BASE + FORMATION_P1);

// Image Data
short int *gImageGraphics               =(short int *)IMAGE_GRAPHICS;
short int *gMissileGraphics             =(short int *)MISSILE_GRAPHICS;
short int *gImageShift                  =(short int *)IMAGE_SHIFT;
unsigned char *gImageHeight             =(unsigned char *)IMAGE_HEIGHTS;
unsigned char * gObjectColors     =(unsigned char *)(DD_BASE + OBJECT_COLORS_RAM);
unsigned char * gObjectColorsNTSC =(unsigned char *)(OBJECT_COLORS_NTSC);


unsigned short int gPodCoolDownRate;
unsigned char gPodFireChance;           // higher the value the higher the chance of shooting
unsigned char gOpenCloseMask;           // ANDed with FRAME to determine open/close speed
unsigned char gLaunchEtype;


// group together a bunch of variables that get zeroed out at the start of each level
// this allows us to use memset to save ROM.
typedef struct {
  unsigned char gInitLevel;
  unsigned char gGunCoolDown;
  unsigned char gStationExplosion;  // points to an active station explosion
  unsigned char gRoundCleared;      // non-zero means round has finished
  unsigned char gFormationAttackPattern;
  unsigned char gFormationShipCount;
  unsigned char gConditionRed;
  unsigned char gStarfieldTopStar;
  unsigned char gPlayerSpeed;
  unsigned char gNewHighScore;
  short int gStarfieldXoffset;
  int  gShipDir;

} ZEROINITLEVEL;

ZEROINITLEVEL zil;

typedef struct
{
  unsigned char state;
  unsigned char center;
  unsigned short int podcooldown[6];
} STATION_STATE;


// sprite data
STATION_STATE gStation[8];
unsigned short int gSpriteState[MAX_SPRITES+1]; // +1 is used for Station Pod collision detection
unsigned char gSpriteImageID[MAX_SPRITES+1];    // +1 is used for Station Pod collision detection
unsigned short int gSpriteX[MAX_SPRITES+1];     // +1 is used for Station Pod collision detection
unsigned short int gSpriteY[MAX_SPRITES+1];     // +1 is used for Station Pod collision detection
unsigned char gSpriteFlickerList1[MAX_SPRITES];
unsigned char gSpriteFlickerList2[MAX_SPRITES];
unsigned char gSpriteColor[MAX_SPRITES];
static unsigned char* gNonFlickeredSprites;
unsigned char * gStationImages =(unsigned char *)(DD_BASE + STATION_IMAGES);

// missile data
unsigned char gMissileState[MAX_MISSILES+1];          // +1 is used for drawing Station Energy Pod
unsigned short int gMissileX[MAX_MISSILES+1];         // +1 is used for drawing Station Energy Pod
unsigned short int gMissileY[MAX_MISSILES+1];         // +1 is used for drawing Station Energy Pod
unsigned char gMissileDirSpeed[MAX_MISSILES];
unsigned char gMissileFlickerList1[MAX_MISSILES];
unsigned char gMissileFlickerList2[MAX_MISSILES];
static unsigned char* gNonFlickeredMissiles;

// Starfield data
short int gStarfieldYoffset;
unsigned char gStarfieldX[MAX_STARS];
unsigned char gStarfieldColor[MAX_STARS];

static void AsteroidDestroyed(int sprite);
static void CheckSampleFinished();
static void ClearGameDatastreams();
static int Collision(int image1_id, int left1, int top1, int sprite_index);
static int ColorConvert(int color);
static unsigned int ConvertToBCD(int value);
static void DeathSequences();
static void DisplayMissiles();
static void DisplaySprites();
static void DisplayStars();
static int DirectionToPlayer(int x, int y, int dir, int object, int turn_mode);
static int DrawMissile(int stream, int index, int firstcall, int color);
static int DrawSprite(int stream, int index, int lumaboost);
static void EndSample();
static void EnemyDestroyed(int sprite);
static void FixSizes();
static void GameOverScan();
static void GameVerticalBlank();
static void IncreaseScore(int points);
static void InitDeathSequence(int sprite, int imageid, unsigned short int x, unsigned short int y, int deadlydebris);
static void Initialize();
static void InitLevel();
static void InitUnusedSprites();
static void MenuOverScan();
static void MenuVerticalBlank();
static void MineDestroyed(int sprite);
static void MissileCollisions();
static void MoveEnemyShips();
static void MoveFormation();
static void MoveMissiles();
static void MoveSpyShip();
static void NewGame();
static int NextUnusedExplosionSprite();
static int NextUnusedEnemySprite();
static int PauseChanged();
static void PlayerDecreaseLives();
static void PlayerDestroyed();
static void PrepScoreRadar();
static void PrepStationGraphics();
static void ProcessJoystick();
static void ProcessMenuJoystick();
static void ProcessSwitches();
//unsigned int Random32();
static unsigned short int ResetTimer(unsigned short int min, unsigned short int range);
static void SetGameDatastreams();
static void ShipCollisions();
static void SpawnEnemyShip(int enemyID, int index, int side);
static void SplashOverScan();
static void SplashVerticalBlank();
static void StationDestroyed();
static void StationMissiles();
static void StationOpenClose();
static void TriggerLoopingSoundEffect(int SoundID, int LoopPoint);
static void TriggerSoundEffect(int SoundID);
static void TriggerSample(int SampleID);
static int TriggerTune(int TuneIDa, int TuneIDb);
static int TurnAwayFromPlayer(int x, int y, /* int object, */ int dir);
static void UpdateSound();

// used to test missile/ball display
//const unsigned char mit[7] = {0, 1, 2, 3, 4, 8, 12};

// speed_dir tables used for sprite movement and station shots, but *NOT* player shots
const signed char speed_dir_x[] =
{
  0,
  14,
  28,
  41,
  52,
  62,
  68,
  73,
  74,
  73,
  68,
  62,
  52,
  41,
  28,
  14,
  0,
  -14,
  -28,
  -41,
  -52,
  -62,
  -68,
  -73,
  -74,
  -73,
  -68,
  -62,
  -52,
  -41,
  -28,
  -14
};

const signed char speed_dir_y[] =
{
  -64,
  -63,
  -60,
  -54,
  -46,
  -36,
  -25,
  -13,
  0,
  12,
  24,
  35,
  45,
  53,
  59,
  62,
  64,
  62,
  59,
  53,
  45,
  35,
  24,
  12,
  0,
  -13,
  -25,
  -36,
  -46,
  -54,
  -60,
  -63
};




// Entry Point
int main()
{
    // Call code here
    switch (FUNC) {
        case RUN_INIT:          Initialize();           break;
        case RUN_MENU_VB:       MenuVerticalBlank();    break;
        case RUN_MENU_OS:       MenuOverScan();         break;
        case RUN_GAME_VB:       GameVerticalBlank();    break;
        case RUN_GAME_OS:       GameOverScan();         break;
        case RUN_SPLASH_VB:     SplashVerticalBlank();  break;
        case RUN_SPLASH_OS:     SplashOverScan();       break;
        case RUN_INIT_LEVEL:    InitLevel();            break;
        default:                                        break;
    }
    return 0;
}

static void Initialize()
{
  int i;
  gSystemIs7800 = VB_TIME;  // 0 = 2600, non-zero = 7800
  
  // in theory this is 4 times faster
  // gave 1.25K of Display Data RAM to C Variable Space
  myMemsetInt((unsigned int *)DD_BASE, 0, (4096-1024-256-64) / 4);
  //    myMemset((unsigned char *)DD_BASE, 0, 4096-1024);
  
  // init all datastreams to increment of 1
  for(i=0;i<=33;i++)
    setIncrement(i,1,0);
  setIncrement(DS_STATION_COLOR,  0, 0);
  
  MODE                = 0x80;            // 0 = Game, 1 = Menu, 0x80 = Splash
  MM_SELECTED_OPTION  = MENU_START_ID;
  MM_LEVEL            = 2;
  MM_TV_TYPE          = NTSC;
  
  gCurrentSample       = 0;  // 0 = no sample
  for(i=0;i<2;i++)
    setNote(i,0);
  setSamplePtr(DD_BASE + 12); //(int)&AUDV0);
  resetWave(0);
}

static void SplashVerticalBlank()
{
  int i;
  int color1;
  int color2;
  
  // MODE is initialized to 0x80
  MODE++;
  
  color1 = color2 = 0;  // blank out logo
  
  if (MODE == 0x81)
  {
    T1TC = 0;           // make sure timer starts at 0
    T1TCR = 1;          // turn on timer
  }
  else if (MODE == 0x82)
  {
    T1TCR = 0;          // turn off timer
    
    if (T1TC < (0x11e8ff + 0x11d329)/2)
      MM_TV_TYPE = NTSC;
    else if (T1TC > (0x11fd2b + 0x11e8ff)/2)
      MM_TV_TYPE = PAL;
    else
      MM_TV_TYPE = SECAM;
    
  }
  else if (MODE >= 0x8c)  // give image time to stabilize before showing logo
  {
    if (MODE == 0x8c)
      TriggerSoundEffect(SOUND_JINGLE);
    color1 = ColorConvert(0x36);
    color2 = ColorConvert(0x96);
  }
  else if (!MODE)
  {
    MODE = 1; // switch to menu
  }
  
  // copy over AtariAge Logo from ROM to RAM
  myMemcpy((unsigned char *)QUEUE + MAIN_MENU_GRAPHICS_RAM + MAIN_MENU_OPTIONS_OFFSET,
           (unsigned char *)SPLASH_SCREEN_GRAPHICS_ROM,
           SPLASH_SCREEN_GFX_SIZE);
  
  for (i=0;i<50;i+=2)
  {
    QUEUE[MAIN_MENU_GRAPHICS_RAM + MAIN_MENU_COLORS_OFFSET + 0 + i] = color1;
    QUEUE[MAIN_MENU_GRAPHICS_RAM + MAIN_MENU_COLORS_OFFSET + 1 + i] = color2;
  }
  
  setPointer(DS_MENU_A,       MAIN_MENU_GRAPHICS_RAM + MAIN_MENU_OPTIONS_OFFSET);
  setPointer(DS_MENU_COLOR,   MAIN_MENU_GRAPHICS_RAM + MAIN_MENU_COLORS_OFFSET);
//  setIncrement(DS_MENU_A,       1, 0);
//  setIncrement(DS_MENU_COLOR,   1, 0);
}

static void SplashOverScan()
{
    // eventually this will allow the fire button to abort the splash screen
    // but only if MODE > 0x82 so the NTSC/PAL detect is finished
  UpdateSound();
}


static void ShowHexInMenu(unsigned int graphic48, unsigned int score)
{
  int digit;
  int r;
  int i;
  int shift;
  shift = 0;
  i = 3;
  do {
    digit = (score & 0x0f) * 8;
    score >>= 4;
    
    for(r=0;r<8;r++)
      QUEUE[graphic48 + 2 + i + r*6] ^= gMenuDigitGfx[digit + r] << shift;
    
    if (shift) i--;
    shift ^= 4;
    
  } while (score);
}

static void ChangeMenuOption(unsigned int graphic48, unsigned int newgraphic, unsigned int columns)
{
  // note when passing graphic48 pass the location of the initial column to change
  
  int r, c;
  for (r=0;r<8;r++)
    for(c=0;c<columns;c++)
      QUEUE[graphic48 + c + r*6] = ((unsigned char *)( newgraphic ))[c + r*columns];
}

static void MenuVerticalBlank()
{
  int i;
  
  // copy over all menu graphics and NTSC colors from Display Data ROM as the RAM is repurposed during the game.
  // this also inits the options to their default values, so we just have to change those that are different.
  myMemcpy((unsigned char *)QUEUE + MAIN_MENU_GRAPHICS_RAM,
           (unsigned char *)MAIN_MENU_GRAPHICS_ROM,
           MAIN_MENU_GFX_SIZE);
  
  if (SWCHB & 0x40) // show debug info if LEFT DIFF = A
  {
//    ShowHexInMenu(MENU_GRAPHICS_OFFSET[MENU_START_ID], gSystemIs7800 ? 0x7800 : 0x2600, T1TC);
#ifdef SHOW_AUDIO_PADDING
    ShowHexInMenu(MENU_GRAPHICS_OFFSET[MENU_START_ID], gPadding);
#endif
    ShowHexInMenu(MENU_GRAPHICS_OFFSET[MENU_LAST_ID], gSystemIs7800 ? 0x7800 : 0x2600);
    ShowHexInMenu(MENU_GRAPHICS_OFFSET[MENU_HIGH_ID], T1TC);
  }
  else
  {
    ShowHexInMenu(MENU_GRAPHICS_OFFSET[MENU_LAST_ID], gLastScore);
    ShowHexInMenu(MENU_GRAPHICS_OFFSET[MENU_HIGH_ID], gHighScore);
  }
  
  
  for(i = 0; i < TITLE_SCREEN_COLOR_SIZE; i++)
    QUEUE[MAIN_MENU_GRAPHICS_RAM + TITLE_SCREEN_COLOR_OFFSET + i] = ColorConvert(QUEUE[MAIN_MENU_GRAPHICS_RAM + TITLE_SCREEN_COLOR_OFFSET + i]);
  
  // fill in white for START, LAST and HIGH score
  for(i=0;i < 3 * 8 * 2;i++)
    QUEUE[MAIN_MENU_GRAPHICS_RAM + START_COLOR_OFFSET + i] = 0x0e;

  switch (MM_QUADRANT)
  {   // Quadrant: [0]=Alpha, 1=Beta, 2=Gamma, 3=Delta
    case 1:
      ChangeMenuOption(MENU_GRAPHICS_OFFSET[MENU_QUADRANT_ID] + 5, OPTION_BETA, 1);
      break;
      
    case 2:
      ChangeMenuOption(MENU_GRAPHICS_OFFSET[MENU_QUADRANT_ID] + 5, OPTION_GAMMA, 1);
      break;
      
    case 3:
      ChangeMenuOption(MENU_GRAPHICS_OFFSET[MENU_QUADRANT_ID] + 5, OPTION_DELTA, 1);
      break;
  }
  
  ShowHexInMenu(MENU_GRAPHICS_OFFSET[MENU_SECTOR_ID], ConvertToBCD(MM_INITIAL_SECTOR+1));
  
  switch (MM_LEVEL)
  {   // skill level: [0]=kids, 1=easy, 2=normal, 3=hard
    case 1:
      ChangeMenuOption(MENU_GRAPHICS_OFFSET[MENU_LEVEL_ID] + 4, OPTION_EASY, 2);
      break;
      
    case 2:
      ChangeMenuOption(MENU_GRAPHICS_OFFSET[MENU_LEVEL_ID] + 3, OPTION_NORMAL, 3);
      break;
      
    case 3:
      ChangeMenuOption(MENU_GRAPHICS_OFFSET[MENU_LEVEL_ID] + 4, OPTION_HARD, 2);
      break;
  }
  
  switch (MM_TV_TYPE)
  {   // TV Type: [0]=NTSC, 1=PAL, 2=SECAM
    case 1:
      ChangeMenuOption(MENU_GRAPHICS_OFFSET[MENU_TV_TYPE_ID] + 4, OPTION_PAL, 2);
      break;
      
    case 2:
      ChangeMenuOption(MENU_GRAPHICS_OFFSET[MENU_TV_TYPE_ID] + 3, OPTION_SECAM, 3);
      break;
  }
  
  // flash selected option
  for(i=0;i<16;i+=2)
  {
    QUEUE[MAIN_MENU_GRAPHICS_RAM + MAIN_MENU_COLORS_OFFSET + (47-8)*2 + MM_SELECTED_OPTION*16 + i + 1] = FRAME;  // visible on NTSC, PAL and SECAM systems
  }
  
  // flash high score if the high score tune is playing
  if (SOUND_VOL[gActiveSound[1]] != 0
      && gActiveSound[1] > TUNE_HIGH_SCORE_A
      && gActiveSound[1] <= TUNE_HIGH_SCORE_B)
  {
    for(i=0;i<16;i+=2)
      QUEUE[MAIN_MENU_GRAPHICS_RAM + MAIN_MENU_COLORS_OFFSET + (47-8)*2 + MENU_HIGH_ID*16 + i + 1] = 255 - FRAME;  // visible on NTSC, PAL and SECAM systems
  }
  
//  // set the datastreams
//  offsetA = FRAME & 1;
////  offsetA = (FRAME & 2) >> 1;
//  offsetB = 1 - offsetA;
//  
//  setPointer(DS_LOGO_A,       MAIN_MENU_GRAPHICS_RAM + 0*LOGO_HEIGHT + offsetA);
//  setPointer(DS_LOGO_B,       MAIN_MENU_GRAPHICS_RAM + 1*LOGO_HEIGHT + offsetB);
//  setPointer(DS_LOGO_C,       MAIN_MENU_GRAPHICS_RAM + 2*LOGO_HEIGHT + offsetA);
//  setPointer(DS_LOGO_D,       MAIN_MENU_GRAPHICS_RAM + 3*LOGO_HEIGHT + offsetB);
//  setPointer(DS_LOGO_E,       MAIN_MENU_GRAPHICS_RAM + 4*LOGO_HEIGHT + offsetA);
//  setPointer(DS_LOGO_F,       MAIN_MENU_GRAPHICS_RAM + 5*LOGO_HEIGHT + offsetB);
//  setPointer(DS_LOGO_G,       MAIN_MENU_GRAPHICS_RAM + 6*LOGO_HEIGHT + offsetA);
//  setPointer(DS_LOGO_H,       MAIN_MENU_GRAPHICS_RAM + 7*LOGO_HEIGHT + offsetB);
//  setPointer(DS_LOGO_I,       MAIN_MENU_GRAPHICS_RAM + 8*LOGO_HEIGHT + offsetA);
//  setPointer(DS_LOGO_J,       MAIN_MENU_GRAPHICS_RAM + 9*LOGO_HEIGHT + offsetB);
//  setPointer(DS_LOGO_K,       MAIN_MENU_GRAPHICS_RAM + 10*LOGO_HEIGHT + offsetA);
//  setPointer(DS_LOGO_L,       MAIN_MENU_GRAPHICS_RAM + 11*LOGO_HEIGHT + offsetB);
//  setPointer(DS_PF2,          MAIN_MENU_GRAPHICS_RAM + 12*LOGO_HEIGHT);
//  setPointer(DS_LOGO_COLORS,  MAIN_MENU_GRAPHICS_RAM + 13*LOGO_HEIGHT);
  setPointer(DS_MENU_CONTROL, MAIN_MENU_GRAPHICS_RAM + MAIN_MENU_CONTROL_OFFSET);
  setPointer(DS_MENU_A,       MAIN_MENU_GRAPHICS_RAM + MAIN_MENU_OPTIONS_OFFSET + 0);
//  setPointer(DS_MENU_B,       MAIN_MENU_GRAPHICS_RAM + MAIN_MENU_OPTIONS_OFFSET + 1);
//  setPointer(DS_MENU_C,       MAIN_MENU_GRAPHICS_RAM + MAIN_MENU_OPTIONS_OFFSET + 2);
//  setPointer(DS_MENU_D,       MAIN_MENU_GRAPHICS_RAM + MAIN_MENU_OPTIONS_OFFSET + 3);
//  setPointer(DS_MENU_E,       MAIN_MENU_GRAPHICS_RAM + MAIN_MENU_OPTIONS_OFFSET + 4);
//  setPointer(DS_MENU_F,       MAIN_MENU_GRAPHICS_RAM + MAIN_MENU_OPTIONS_OFFSET + 5);
  setPointer(DS_MENU_COLOR,   MAIN_MENU_GRAPHICS_RAM + MAIN_MENU_COLORS_OFFSET);
  
  
//  setIncrement(DS_LOGO_A,       2, 0);
//  setIncrement(DS_LOGO_B,       2, 0);
//  setIncrement(DS_LOGO_C,       2, 0);
//  setIncrement(DS_LOGO_D,       2, 0);
//  setIncrement(DS_LOGO_E,       2, 0);
//  setIncrement(DS_LOGO_F,       2, 0);
//  setIncrement(DS_LOGO_G,       2, 0);
//  setIncrement(DS_LOGO_H,       2, 0);
//  setIncrement(DS_LOGO_I,       2, 0);
//  setIncrement(DS_LOGO_J,       2, 0);
//  setIncrement(DS_LOGO_K,       2, 0);
//  setIncrement(DS_LOGO_L,       2, 0);
//  setIncrement(DS_PF2,          1, 0);
//  setIncrement(DS_LOGO_COLORS,  1, 0);
//dgs  setIncrement(DS_MENU_CONTROL, 1, 0);
//dgs  setIncrement(DS_MENU_A,       1, 0);
//  setIncrement(DS_MENU_B,       6, 0);
//  setIncrement(DS_MENU_C,       6, 0);
//  setIncrement(DS_MENU_D,       6, 0);
//  setIncrement(DS_MENU_E,       6, 0);
//  setIncrement(DS_MENU_F,       6, 0);
//dgs  setIncrement(DS_MENU_COLOR,   1, 0);
  CheckSampleFinished(); // check sample as last thing in VB and first in OS
}

static void MenuOverScan()
{
  CheckSampleFinished(); // check sample as last thing in VB and first in OS
  ProcessSwitches();    // processes SELECT and RESET
  if (MODE == 1)            // if MODE = 0, then we're about to start the game
  {
    ProcessMenuJoystick();
  }
  zil.gConditionRed = 0;  // make sure condition red sound doesn't get triggered
  UpdateSound();
}

static void GameVerticalBlank()
{
  ClearGameDatastreams();
  if (zil.gInitLevel)
  {
    gJumpDataStream[0] = INIT_LEVEL_KERNEL;
  }
  else
  {
    if (!gPaused)
    {
      DisplaySprites();
      DisplayMissiles();
      FixSizes();
    }
    DisplayStars();
  }
  PrepScoreRadar();       // uses value in gJumpDataStream[0] to know if level data is valid
  SetGameDatastreams();
  CheckSampleFinished();  // check sample as last thing in VB and first in OS
}

static void GameOverScan()
{
  int i;
  CheckSampleFinished(); // check sample as last thing in VB and first in OS
  ProcessSwitches();
  if (MODE == 0) // if MODE = 1, then we're about to go to the menu
  {
    if (PauseChanged())
      gPaused = !gPaused;
    
    if (gPaused)
    {
//      for (i=0;i<2;i++)
//      {
//        AUDC[i] = 0;
//        AUDV[i] = 0;
//        AUDF[i] = 0;
//        gActiveSound[i] = 0;  // was missing this
//      }
      
      for (i=AUDIO_QUIET;i<AUDIO_QUIET+6;i++)  // saves 36 bytes
        QUEUE[i] = 0;
      
      gActiveSound[0] = gActiveSound[1] = 0;
      
    }
    else
    {
      DeathSequences();
      InitUnusedSprites();
      if (gSpriteState[PLAYER_SPRITE] & SPS_DEADLY_TO_ENEMY)  // true if player is still alive
      {
        MissileCollisions();   // done after shots were drawn so collisions can be seen
        ShipCollisions();
      }
      ProcessJoystick();
      if (zil.gPlayerSpeed == 128 && !zil.gRoundCleared && !gRoundStart)
      {
#ifdef RIGHT_DIFFICULTY_SAFE
        if (!(SWCHB & 0x80))   // Right Difficulty A = stationary ship
#endif
        {
          StationMissiles();
          StationOpenClose(); // zeros & updates gPriorFrameEnemyMissileOnscreen
          MoveFormation();    //         updates gPriorFrameEnemyMissileOnscreen
          MoveEnemyShips();   //         updates gPriorFrameEnemyMissileOnscreen
          MoveSpyShip();      //         updates gPriorFrameEnemyMissileOnscreen
        }
      }
      MoveMissiles();
      PrepStationGraphics();
    }
    UpdateSound();
  }
}

const unsigned char * const * const levels[] =
{
  alpha_levels,
  beta_levels,
  gamma_levels,
  delta_levels
};

const unsigned char sector_counts[] =
{
  sizeof(alpha_levels) / sizeof(alpha_levels[0]),
  sizeof(beta_levels) / sizeof(beta_levels[0]),
  sizeof(gamma_levels) / sizeof(gamma_levels[0]),
  4 //sizeof(delta_levels) / sizeof(delta_levels[0])
};

const unsigned char sector_repeat[] =
{
  6,  // repeat the last 6 levels
  6,  // repeat the last 6 levels
  sizeof(gamma_levels) / sizeof(gamma_levels[0]),
  4 //sizeof(delta_levels) / sizeof(delta_levels[0])
};

// pods are numbered 0-5 in this arrangement around the station cores:
//  0     0 1
// 5 1   5   2  <- Vertical
// 4 2    4 3
//  3  <---------- Horizontal
const unsigned char pod_offset_x[] =
{
  10, // Horizontal Pod 0
  20, // Horizontal Pod 1
  20, // Horizontal Pod 2
  10, // Horizontal Pod 3
  0, // Horizontal Pod 4
  0, // Horizontal Pod 5
  10, // Horizontal Energy Pod
  6, // Vertical Pod 0
  18, // Vertical Pod 1
  24, // Vertical Pod 2
  18, // Vertical Pod 3
  6, // Vertical Pod 4
  0, // Vertical Pod 5
  14  // Vertical Energy Pod
};

const unsigned char pod_offset_y[] =
{
  0, // Horizontal Pod 0
  13, // Horizontal Pod 1
  37, // Horizontal Pod 2
  50, // Horizontal Pod 3
  37, // Horizontal Pod 4
  13, // Horizontal Pod 5
  29, // Horizontal Energy Pod
  0, // Vertical Pod 0
  0, // Vertical Pod 1
  21, // Vertical Pod 2
  42, // Vertical Pod 3
  42, // Vertical Pod 4
  21, // Vertical Pod 5
  22  // Vertical Energy Pod
};

// pods are numbered 0-5 in this arrangement around the station cores:
//  0     0 1
// 5 1   5   2  <- Vertical
// 4 2    4 3
//  3  <---------- Horizontal
const signed char pod_shot_offset_x[] =
{
  10 + 3,   // Horizontal Pod 0
  20 + 8,   // Horizontal Pod 1
  20 + 8,   // Horizontal Pod 2
  10 + 3,   // Horizontal Pod 3
  0 - 2,    // Horizontal Pod 4
  0 - 2,    // Horizontal Pod 5
  10,       // Horizontal Energy Pod  NOT USED
  6 + 0,    // Vertical Pod 0
  18 + 6,   // Vertical Pod 1
  24 + 8,   // Vertical Pod 2
  18 + 6,   // Vertical Pod 3
  6 + 0,    // Vertical Pod 4
  0 - 2,    // Vertical Pod 5
  14        // Vertical Energy Pod    NOT USED
};

const signed char pod_shot_offset_y[] =
{
  0 - 4,    // Horizontal Pod 0
  13 + 2,   // Horizontal Pod 1
  37 + 9,   // Horizontal Pod 2
  50 + 15,  // Horizontal Pod 3
  37 + 9,   // Horizontal Pod 4
  13 + 2,   // Horizontal Pod 5
  0,        // Horizontal Energy Pod  NOT USED
  0 - 2,    // Vertical Pod 0
  0 - 2,    // Vertical Pod 1
  21 + 5,   // Vertical Pod 2
  42 + 13,  // Vertical Pod 3
  42 + 13,  // Vertical Pod 4
  21 + 5,   // Vertical Pod 5
  22        // Vertical Energy Pod    NOT USED
};

const unsigned char pod_shot_init_dir[] =
{
  0,        // Horizontal Pod 0
  8-3,      // Horizontal Pod 1
  8+3,      // Horizontal Pod 2
  16,       // Horizontal Pod 3
  24-3,     // Horizontal Pod 4
  24+3,     // Horizontal Pod 5
  0,        // Horizontal Energy Pod  NOT USED
  32-3,     // Vertical Pod 0
  0+3,      // Vertical Pod 1
  8,        // Vertical Pod 2
  16-3,     // Vertical Pod 3
  16+3,     // Vertical Pod 4
  24,       // Vertical Pod 5
  0         // Vertical Energy Pod    NOT USED
};

const unsigned char pod_shape[] =
{
  POD_ID,                     // Horizontal Pod 0
  POD_ID,                     // Horizontal Pod 1
  POD_ID,                     // Horizontal Pod 2
  POD_ID,                     // Horizontal Pod 3
  POD_ID,                     // Horizontal Pod 4
  POD_ID,                     // Horizontal Pod 5
  CORE_HORIZONTAL_ID,         // Horizontal Energy Pod
  POD_ID,                     // Vertical Pod 0
  POD_ID,                     // Vertical Pod 1
  POD_ID,                     // Vertical Pod 2
  POD_ID,                     // Vertical Pod 3
  POD_ID,                     // Vertical Pod 4
  POD_ID,                     // Vertical Pod 5
  CORE_VERTICAL_ID            // Vertical Energy Pod
};

static unsigned int ConvertToBCD(int value)
{
  unsigned int bcd;
  bcd = 0;
  
  while(value >= 10000)
  {
    bcd += 0x10000;
    value -= 10000;
  }
  while(value >= 1000)
  {
    bcd += 0x1000;
    value -= 1000;
  }
  while(value >= 100)
  {
    bcd += 0x100;
    value -= 100;
  }
  while(value >= 10)
  {
    bcd += 0x10;
    value -= 10;
  }
  bcd += value;
  
  return bcd;
}

static const unsigned char * SetPlayerPosition()
{
  int i;
  int l;
  const unsigned char *level;
  
  i = gSector;
  
  while (i >= sector_counts[MM_QUADRANT])
    i -= sector_repeat[MM_QUADRANT];
  
  level = levels[MM_QUADRANT][i];
  
  // position the player at start of level
  l = 0;
  gSpriteState[PLAYER_SPRITE]     = SPS_DEADLY_TO_ENEMY;
  gSpriteImageID[PLAYER_SPRITE]   = SHIP_ID;
  gSpriteColor[PLAYER_SPRITE]     = gObjectColors[COLOR_SHIP];
  gSpriteX[PLAYER_SPRITE]         = (level[l++] * 2) << X_FRACTION_BITS;
  gSpriteY[PLAYER_SPRITE]         = (level[l++] * 4) << Y_FRACTION_BITS;
  zil.gShipDir                    = 0;
  zil.gPlayerSpeed                = 0;
  gPriorFrameStationOnscreen      = 0;
  zil.gConditionRed               = 0;
  gPriorFrameEnemyMissileOnscreen = 0;
  gNotSoChatty                    = 0;
  gIdleConditionRedTimer = gIdleConditionRedMax;  // reset idle timer
  
  gWindowX = gSpriteX[PLAYER_SPRITE] - (80 << X_FRACTION_BITS);
  gWindowY = gSpriteY[PLAYER_SPRITE] - (((DATASTREAM_SIZE - SHIP_HEIGHT) /2) << Y_FRACTION_BITS);

  // used in InitLevel to load the station data
  return level;
}

static void InitLevel()
{
  int i;
  int s;
  int l;
  int x;
  int y;
  int a;
  int c;
  int p;
  int o;
  const unsigned char *level;
  
  gSectorBCD = ConvertToBCD(gSector + 1);
  
//  i = gSector;
//  
//  while (i >= sector_counts[MM_QUADRANT])
//    i -= sector_repeat[MM_QUADRANT];
//  
//  level = levels[MM_QUADRANT][i];
  
//  zil.gInitLevel = 0;
//  zil.gGunCoolDown = 0;
//  zil.gStationExplosion = 0;
//  zil.gRoundCleared = 0;
//  zil.gFormationAttackPattern = 0;
//  zil.gFormationShipCount = 0;
//  zil.gConditionRed = 0;
  myMemset((unsigned char *)&zil, 0, sizeof(zil));
  
  
  if (gSector < 32)
    s = gSector;
  else
    s = 31; // max difficulty at sector 32
  
  // calculate difficulty ramp up for I & P-Type
  
  gEnemyMax[0] = gEnemyMax[1] = MM_LEVEL;
  if (gEnemyMax[1] ==0) gEnemyMax[1] = 1;       // for kids level only use Intercepters
  gEnemySpeed     = 80 + 20 * MM_LEVEL + s * 5; // note speed 128 = player's ship
// NOTE - MinTime and RangeTime are currently the same, will probably change as the rampup is tweaked
  gEnemyMinTime       = 60 - 5 * MM_LEVEL - s/2;
  gEnemyRangeTime     = 60 - 5 * MM_LEVEL - s/2;    // RangeTime = MaxTime - MinTime
  gEnemyTimer[0]      = ResetTimer(gEnemyMinTime, gEnemyRangeTime);  // 0 = P-Type
  gEnemyTimer[1]      = ResetTimer(gEnemyMinTime, gEnemyRangeTime);  // 1 = I-Type

  // calculate difficulty ramp up for Formations
  gFormationMinTime   = 1200 - 150 * MM_LEVEL - 10 * s;
  gFormationRangeTime = 600 - 150 * MM_LEVEL - 10 * s;   // RangeTime = MaxTime - MinTime
  gFormationTimer     = ResetTimer(gFormationMinTime, gFormationRangeTime);
  gFormationSpeed     = 130 + 5 * MM_LEVEL + 1 * s;
  if      (gSector == 0)  gFormationMask = 0;  // only I-Type on first level
  else if (gSector == 1)  gFormationMask = 3;  // I-Type and P-Type on second level
  else                    gFormationMask = 7;
  
  // spy ship should show up after 3 formations for sectors 0 & 1
  //                               2 formations for sectors 2 & 3
  //                               1 formation  for sectors 4 & 5
  //                               0 formations for 6+
  if (gSector >= 6)
    gFormationSpyCountdown = 0;
  else
    gFormationSpyCountdown = 3 - (gSector / 2);
  
  // calculate difficulty ramp up for Spy Ship
  gSpyMinTime         = 1000 - 150 * MM_LEVEL - 10 * s;
  gSpyRangeTime       = 500 - 100 * MM_LEVEL - 10 * s;   // RangeTime = MaxTime - MinTime
  gSpyTimer           = ResetTimer(gSpyMinTime, gSpyRangeTime);
  gSpySpeed           = 140 + 5 * MM_LEVEL + 1 * s;                      // slightly faster than player
  
  // set idle timer to
  //  kids    = 60 seconds
  //  easy    = 45 seconds
  //  normal  = 30 seconds
  //  hard    = 15 seconds
  gIdleConditionRedMax = gIdleConditionRedTimer = (4-MM_LEVEL) * 15*60;
  
  gNonFlickeredSprites = 0;
  for (i=0;i<MAX_SPRITES;i++)
  {
    gSpriteState[i] = 0;
    gSpriteFlickerList1[i] = i;       // init flicker list
  }
  gSpriteState[MAX_SPRITES] = 0;      // used for collision detection with station pods
  
  gNonFlickeredMissiles = 0;
  for (i=0;i<MAX_MISSILES;i++)
  {
    gMissileState[i] = 0;
    gMissileFlickerList1[i] = i;       // init flicker list
  }
  
  // create new random starfield
//  zil.gStarfieldXoffset = 0;  // now in the zil memset
  gStarfieldYoffset = 2 << Y_FRACTION_BITS;
//  zil.gStarfieldTopStar = 0;  // now in the zil memset
  for (i=0;i<MAX_STARS;i++)
  {
    gStarfieldX[i] = getRandom32() >> 24; // random 0-255
    if (MM_TV_TYPE == 2)
      gStarfieldColor[i] = ((14 * (getRandom32() & 0xff)) >> 8) + 2; // if SECAM use random color 2-15
    else
      gStarfieldColor[i] = ((getRandom32() & 0xF700) >> 8) + 0x08;   // random color with luma 8-15
  }
  
  level = SetPlayerPosition();

  // position the player
//  l = 0;
//  gSpriteState[PLAYER_SPRITE]     = SPS_DEADLY_TO_ENEMY;
//  gSpriteImageID[PLAYER_SPRITE]   = SHIP_ID;
//  gSpriteColor[PLAYER_SPRITE]     = gObjectColors[COLOR_SHIP];
//  gSpriteX[PLAYER_SPRITE]         = (level[l++] * 2) << X_FRACTION_BITS;    //80;
//  gSpriteY[PLAYER_SPRITE]         = (level[l++] * 4) << Y_FRACTION_BITS;    //(DATASTREAM_SIZE - SHIP_HEIGHT) /2;
//  //zil.gShipDir                        = 0;  // now in the zil memset
//  //zil.gPlayerSpeed                    = 0;  // now in the zil memset
//  
//  gWindowX = gSpriteX[PLAYER_SPRITE] - (80 << X_FRACTION_BITS);
//  gWindowY = gSpriteY[PLAYER_SPRITE] - (((DATASTREAM_SIZE - SHIP_HEIGHT) /2) << Y_FRACTION_BITS);
  
  l = 2;
  
  
  o = MM_LEVEL * gSector; // always 0 for kids
  // level  KIDS  EASY  NORMAL  HARD
  //  1     0     0     0       0 <- always open for all difficulty levels
  //  2     0     1     2       3 <- start to open/close for Easy, Normal, and Hard
  //  3     0     2     4       6 <- start E-Type on Hard
  //  4     0     3     6       9 <- start E-Type on Normal
  //  5     0     4     8       12 <- faster open/close on Hard
  //  6     0     5     10      15 <- start E-Type on Easy
  //  7     0     6     12      18 <- faster open/close on Normal
  //  ...
  //  12    0     11    22      33 <- faster open/close on Easy
  
  if      (o == 0)  { s = SCS_ALWAYSOPEN | SCS_FULLY_OPEN;  gLaunchEtype = 0; }
  else if (o <  5)  { s = SCS_OPENING + SCS_FULLY_CLOSED;   gLaunchEtype = 0; }
  else              { s = SCS_OPENING + SCS_FULLY_CLOSED;   gLaunchEtype = 1; }

  if (o < 11)       gOpenCloseMask = 15;  // open/close rate
  else              gOpenCloseMask = 7;  // faster open/close rate

  // init station state
  for(i=0;i<8;i++)
  {
    gStation[i].state = 0;
    for(p=0;p<6;p++)
      gStation[i].podcooldown[p] = 0xffff;
    gStation[i].center = s;
  }

// used when 4 stations can shoot per frame
//  i = 25 +                          // 25/255 = minimum 10% chance a cooled down pod will fire
//      ((MM_LEVEL * gSector)>>2);    // increased based on difficulty level and level
//  gPodCoolDownRate = MM_LEVEL * (0x80 + gSector);
  
// used when 2 stations can shoot per frame
  i = 50 +                          // 50/255 = minimum 20% chance a cooled down pod will fire
  ((MM_LEVEL * gSector)>>2);    // increased based on difficulty level and level
  gPodFireChance = (i > 255) ? 255 : i;
  gPodCoolDownRate = (MM_LEVEL * (0x80 + gSector)) * 2;
  
// used when 1 station can shoot per frame
//  i = 100 +                          // 50/255 = minimum 40% chance a cooled down pod will fire
//  ((MM_LEVEL * gSector)>>2);    // increased based on difficulty level and level
//  gPodFireChance = (i > 255) ? 255 : i;
//  gPodCoolDownRate = (MM_LEVEL * (0x80 + gSector)) * 4; // 1 station per frame

  // position stations
  myMemset(gStationImages, 0, STATION_HEIGHT*8);
  s = level[l++]; // station count
  o = level[l++]; // orientation
  for (i=0;i<s;i++)
  {
    gSpriteState[i]     = SPS_DEADLY_TO_PLAYER | SPS_4X_SPRITE;
    gStation[i].state   = STS_ACTIVE_PODS; // activates all 6 pods
//    gSpriteImageID[i]   = STATION_HORIZONTAL_ID + (level[l] >> VS_SHIFT);
    gSpriteImageID[i]   = o & 1;  // STATION_HORIZONTAL_ID = 0, STATION_VERTIAL_ID = 1
    o >>= 1;
    gSpriteColor[i]     = gObjectColors[COLOR_STATION];
    gSpriteX[i]         = (level[l++] * 2) << X_FRACTION_BITS;
    gSpriteY[i]         = (level[l++] * 4)<< Y_FRACTION_BITS;
    // not needed, station images are prepped every frame in PrepStationGraphics()
//    myMemcpy((unsigned char *)DD_BASE + STATION_IMAGES + i*STATION_HEIGHT,
//             (unsigned char *)0 + gImageGraphics[gSpriteImageID[i]],
//             gImageHeight[gSpriteImageID[i]]);
  }
  
  // visually show pod & energy core collision targets for station 0
//  s = gSpriteImageID[0] * 7;
//  for(i=0;i<7;i++)
//  {
//    gSpriteState[ETYPE_SPRITE + i]    = SPS_DEADLY_TO_PLAYER;
//    gSpriteX[ETYPE_SPRITE + i]        = gSpriteX[0] + (pod_offset_x[s + i] << X_FRACTION_BITS);
//    gSpriteY[ETYPE_SPRITE + i]        = gSpriteY[0] + (pod_offset_y[s + i] << Y_FRACTION_BITS);
//    gSpriteColor[ETYPE_SPRITE + i]    = 0x0d;     // white
//    gSpriteImageID[ETYPE_SPRITE + i]  = pod_shape[s + i];
//  }

  // randomly place asteroids and mines so they don't overlap
  a = FIRST_STATIONARY_SPRITE;
  for (l=0;l<32;l++)
  {
    //        for (i=0; i < d; i++)
    for (i=0; i <= MM_LEVEL; i++)
    {
      do
      {
        x = (( 504 * (getRandom32() & 0xff)) >> 8) << X_FRACTION_BITS;
        y = ((l * 32) + (getRandom32() & 15)) << Y_FRACTION_BITS;      // max stationary height is 15, so object should never straddle 2 vertical zones
        c = 0;  // reset collision
        
        // player overlap check goes here
        // make sure to add some buffer space around player's position
        if (y + (15 << Y_FRACTION_BITS) >= gSpriteY[PLAYER_SPRITE] - (PLAYER_BUFFER_SPACE << Y_FRACTION_BITS))          // 15 = height of mine
          if (y <= gSpriteY[PLAYER_SPRITE] + ((PLAYER_BUFFER_SPACE + 17) << Y_FRACTION_BITS))                         // 17 = height of player's ship
            if (x + (8 << X_FRACTION_BITS) >= gSpriteX[PLAYER_SPRITE] - (PLAYER_BUFFER_SPACE << X_FRACTION_BITS))   // 8 = width of mine
              if (x <= gSpriteX[PLAYER_SPRITE] + ((9 + PLAYER_BUFFER_SPACE) << X_FRACTION_BITS))                  // 9 = width of ship
                c = 1;
        
        // stationary overlap check, only checking for X overlap
        if (c == 0)
          for (s=0;s<i;s++)
          {
            if (x + (8 << X_FRACTION_BITS) >= gSpriteX[a+s])
              if (x <= gSpriteX[a+s] + (8 << X_FRACTION_BITS))
                c = 1;
          }
        
        // station overlap check
        if (c == 0)
          for (s=0;s<8;s++)
          {
            if (gSpriteState[s])
            {
              if (y + (15 << Y_FRACTION_BITS) >= gSpriteY[s])             // 15 = height of mine
                if (y <= gSpriteY[s] + (65 << Y_FRACTION_BITS))         // 65 = height of horizontal station
                  if (x + (8 << X_FRACTION_BITS) >= gSpriteX[s])      // 8 = width of mine
                    if (x <= gSpriteX[s] + (32 << X_FRACTION_BITS)) // 32 = width of station
                      c = 1;
            }
          }
      } while (c);
      gSpriteState[a+i]     = SPS_DEADLY_TO_PLAYER | SPS_DEADLY_TO_ENEMY;
      gSpriteX[a+i]         = x;
      gSpriteY[a+i]         = y;
      gSpriteImageID[a+i]   = (getRandom32() & 3) + ASTEROID_ID;
      gSpriteColor[a+i]     = (gSpriteImageID[a+i] == MINE_ID) ? gObjectColors[COLOR_MINE] : gObjectColors[COLOR_ASTEROID];
    }
    a += STATIONARY_DENSITY;
  }
  
  setPointer(33, SCORE_DATASTREAM);
}

static int PauseChanged()
{
  int changed = 0;
  unsigned int state = (SWCHB & 8);
  
  if (gPaused)
  {
    // firebutton can unpause the game
    if (!(INPT4 & 0x80))
    {
      MM_FIRE_DOWN = 1;
    }
    else if (MM_FIRE_DOWN)
    {
      MM_FIRE_DOWN = 0;
      changed = 1;
    }
  }
  
  if (gSystemIs7800)
  {
    if (state) // color selected, pause not pressed
    {
      if (gColorSwitchState)
      {
        changed = 1;
      }
      gColorSwitchState = 0;
    }
    else
      gColorSwitchState = 1; // bw selected, pause pressed
  }
  else
  {
    if (gColorSwitchState != state)
    {
      changed = 1;
    }
    gColorSwitchState = state;
  }
  return changed;
}

static void ProcessSwitches()
{
  FRAME++;
  getRandom32();                                                            // mix up the random numbers
  if (!(SWCHB & 2))               MODE = 1;     // show the main menu
  else if (!(SWCHB & 1))          NewGame();        // start new game
}

// already in defines_cdf.h as getRandom32()
//unsigned int Random32()
//{
//  // using a 32-bit Galois LFSR as a psuedo random number generator.
//  // http://en.wikipedia.org/wiki/Linear_feedback_shift_register#Galois_LFSRs
//  
//  static unsigned int random = 0x02468ace;
//  
//  return random = (random >> 1) ^ (unsigned int)(-(random & 1u) & 0xd0000001u);
//}


#if 0 // 0 = use merged PAL & SECAM color tables, 1 = use seperate tables
      // surpisingly, the merged tables used 4 bytes more ROM...
const unsigned char NTSCtoPAL[16] =
{
  // SeaGtGruff's conversion values from this post at AtariAge:
  // http://atariage.com/forums/topic/165424-modify-colour-palette/?p=2043124
  0x00, 0x20, 0x20, 0x40, 0x60, 0x80, 0xa0, 0xc0,
  0xd0, 0xb0, 0x90, 0x70, 0x50, 0x30, 0x30, 0x20
};

const unsigned char NTSCtoSECAM[16] =
{
  0x0e,   // 0 = white            e = white
  0x0c,   // 1 = yellow           c = yellow
  0x0c,   // 2 = yellowish orange c = yellow
  0x04,   // 3 = reddish orange   4 = red
  0x04,   // 4 = red              4 = red
  0x06,   // 5 = reddish purple   6 = purple
  0x06,   // 6 = purple           6 = purple
  0x06,   // 7 = bluish purple    6 = purple
  0x02,   // 8 = blue             2 = blue
  0x02,   // 9 = blue             2 = blue
  0x0a,   // a = bluish cyan      a = cyan
  0x0a,   // b = cyan             a = cyan
  0x08,   // c = green            8 = green
  0x08,   // d = green            8 = green
  0x08,   // e = olive green      8 = green
  0x0c    // f = brown            c = yellow
};

//int ColorPAL(int color)
//{
//  return NTSCtoPAL[color>>4] +    // chroma value
//  (color & 0x0f);          // luma value
//}
//
//int ColorSECAM(int color)
//{
//  if (color < 2)
//    return 0;   // return black for 0 or 1
//  else
//    return NTSCtoSECAM[color>>4];
//}

static int ColorConvert(int color)
{
  if (MM_TV_TYPE == PAL)
  {
//    return ColorPAL(color);
    return NTSCtoPAL[color>>4] +    // chroma value
           (color & 0x0f);          // luma value
  }
  else if (MM_TV_TYPE == SECAM)
  {
//    return ColorSECAM(color);
    if (color < 2)
      return 0;   // return black for 0 or 1
    else
      return NTSCtoSECAM[color>>4];
  }
  else
    return color;
}

#else
const unsigned char NTSCtoPALSECAM[16] =
{
  0x0e,   // 0 = white            e = white
  0x2c,   // 1 = yellow           c = yellow
  0x2c,   // 2 = yellowish orange c = yellow
  0x44,   // 3 = reddish orange   4 = red
  0x64,   // 4 = red              4 = red
  0x86,   // 5 = reddish purple   6 = purple
  0xa6,   // 6 = purple           6 = purple
  0xc6,   // 7 = bluish purple    6 = purple
  0xd2,   // 8 = blue             2 = blue
  0xb2,   // 9 = blue             2 = blue
  0x9a,   // a = bluish cyan      a = cyan
  0x7a,   // b = cyan             a = cyan
  0x58,   // c = green            8 = green
  0x38,   // d = green            8 = green
  0x38,   // e = olive green      8 = green
  0x2c    // f = brown            c = yellow
};

//int ColorPAL(int color)
//{
//  return (NTSCtoPALSECAM[color>>4] & 0xf0) +    // chroma value
//         (color & 0x0f);                        // luma value
//}
//
//int ColorSECAM(int color)
//{
//  if (color < 2)
//    return 0;   // return black for 0 or 1
//  else
//    return NTSCtoPALSECAM[color>>4];            // SECAM only uses luma so upper nybble is ignored
//}

static int ColorConvert(int color)
{
  unsigned char basecol = NTSCtoPALSECAM[color>>4];
  
  if (MM_TV_TYPE == PAL)
  {
    return (basecol & 0xf0) +    // chroma value
           (color & 0x0f);       // luma value
  }
  else if (MM_TV_TYPE == SECAM)
  {
    // SECAM only uses luma so upper nybble is ignored
    return (color < 2) ? 0 : basecol; // return black for 0 or 1
  }
  else
    return color;
}

#endif // merged PAL & SECAM data

static void TriggerSoundEffect(int SoundID)
{
  int channel;
  
  // sample could be played at any time on channel 0, so try channel 1 first
  if (SOUND_VOL[gActiveSound[1]] == 0)                              channel = 1;
  else if (gCurrentSample == 0 && SOUND_VOL[gActiveSound[0]] == 0)  channel = 0;
  else if (SoundID > gActiveSound[1])                               channel = 1;
  else if (gCurrentSample == 0 && SoundID > gActiveSound[0])        channel = 0;
  else                                                            return;
  
  gActiveSound[channel] = SoundID;
  gSoundDuration[channel] = SOUND_FREQ[SoundID] >> 5;
}

static void TriggerLoopingSoundEffect(int SoundID, int LoopPoint)
{
  if (gActiveSound[1] > LoopPoint && gActiveSound[1] <= SoundID)
    return;
  gActiveSound[1] = SoundID;
  gSoundDuration[1] = SOUND_FREQ[SoundID] >> 5;
}

static int TriggerTune(int TuneIDa, int TuneIDb)
{
  // make sure any sample is finished before triggering a tune
  if (gCurrentSample == 0)
  {
    gActiveSound[0]   = TuneIDa;
    gActiveSound[1]   = TuneIDb;
    gSoundDuration[0] = SOUND_FREQ[TuneIDa] >> 5;
    gSoundDuration[1] = SOUND_FREQ[TuneIDb] >> 5;
    return 1;
  }
  return 0;
}


static void UpdateSound()
{
  int channel;
  unsigned char vol;
  
  if (zil.gConditionRed)
    TriggerLoopingSoundEffect(SOUND_CONDITION_RED_ID, SOUND_CONDITION_RED_LOOP);
  
  for (channel = (gCurrentSample ? 1 : 0);channel < 2; channel++)
  {
    vol = SOUND_VOL[gActiveSound[channel]];
    // due to how sample playback works, we need both nybbles set to the same
    // value when using VOICE 0 for sound effects
    if (channel == 0)
    {
      AUDV0 = (vol & 0x0f) * 0x11;
//      AUDC0 = vol >> 4;
//      AUDF0 = SOUND_FREQ[gActiveSound[channel]];
    }
    else
    {
      AUDV1 = vol & 0x0f;
      if (gCurrentSample)
        AUDV1 = (AUDV1 + 3) / 4;
      
//      AUDC1 = vol >> 4;
//      AUDF1 = SOUND_FREQ[gActiveSound[channel]];
    }
    AUDC[channel] = vol >> 4;
    AUDF[channel] = SOUND_FREQ[gActiveSound[channel]];
    
    
    if (vol)  // if AUDV and AUDC are non-zero then sound effect is active
    {
      if (gSoundDuration[channel])
      {
        gSoundDuration[channel]--;
      }
      else
      {
        gActiveSound[channel]--;
        gSoundDuration[channel] = SOUND_FREQ[gActiveSound[channel]] >> 5;
      }
    }
  }
}

static void TriggerSample(int SampleID)
{
  // don't abort a higher priority sample
  if (SampleID < gCurrentSample)
    return;
  
  //turn off any active sound effect on channel 0
  gActiveSound[0] = 0;
  AUDC0 = 0;
  AUDF0 = 0;

  // prep to play sample on channel 0
  gCurrentSample = SampleID;
  resetWave(0);
  setSamplePtr(SAMPLES[gCurrentSample-1]);
  setNote(0, SAMPLE_RATES[gCurrentSample-1] * 256);
}

static void EndSample()
{
  gCurrentSample = 0;  // 0 = no sample
  setNote(0, 0);
  resetWave(0);
  setSamplePtr(DD_BASE + 12); //(int)&AUDV0);
}

static void CheckSampleFinished()
{
#ifdef SHOW_AUDIO_PADDING
  int pad;
#endif
  
    if (gCurrentSample)
    {
      if ((getWavePtr(0) >> 21) > SAMPLE_SIZES[gCurrentSample-1])
      {
#ifdef SHOW_AUDIO_PADDING
        pad = (getWavePtr(0) >> 21) - SAMPLE_SIZES[gCurrentSample-1];
        if (pad > gPadding)
          gPadding = pad;
#endif
        
        if (gCurrentSample == SAMPLE_ALERT_ALERT && gAlertRepeat)
        {
          TriggerSample(SAMPLE_ALERT_ALERT);
          gAlertRepeat = 0;
        }
        else
        {
          EndSample();
        }
      }
    }
}

static void ProcessMenuJoystick()
{
  unsigned char orig_quadrant;
  unsigned char orig_sector;
  unsigned char orig_level;
  
  
  orig_quadrant = MM_QUADRANT;
  orig_sector   = MM_INITIAL_SECTOR;
  orig_level    = MM_LEVEL;
  
  if (!(INPT4 & 0x80))	// firebutton down
  {
    if (!gIgnoreFireInMenu)
      MM_FIRE_DOWN = 1;
  }
  else
  {
    if (MM_FIRE_DOWN)	// if fire button was released
    {
      MM_FIRE_DOWN = 0;
      if (MM_SELECTED_OPTION == MENU_START_ID)
        NewGame();  // start game when firebutton released
    }
    gIgnoreFireInMenu = 0;
  }
  
  // joystick timer controls auto-changing of values when fire is held
  if (((SWCHA & 0xF0) == 0xF0) && (!MM_FIRE_DOWN)) // if nothing is down
  {
    MM_JOYSTICK_TIMER = 0;
  }
  else
  {
    MM_JOYSTICK_TIMER++;
    if (MM_JOYSTICK_TIMER == 16)
      MM_JOYSTICK_TIMER = 0;
  }
  if (MM_JOYSTICK_TIMER == 1)
  {
    if (!(SWCHA & 0x10) && (!MM_FIRE_DOWN))			// if joystick up and no fire button then change selected option
    {
      if (MM_SELECTED_OPTION ==  MENU_FIRST_OPTION_ID)
        MM_SELECTED_OPTION = MENU_START_ID;
      else
        MM_SELECTED_OPTION--;
      
      return;
    }
    if (!(SWCHA & 0x20) && (!MM_FIRE_DOWN))	// if joystick down and no firebutton then change selected option
    {
      if (MM_SELECTED_OPTION == MENU_START_ID)
        MM_SELECTED_OPTION = MENU_FIRST_OPTION_ID;
      else
        MM_SELECTED_OPTION++;
      
      return;
    }
    if (!(SWCHA & 0x40))	// joystick left
    {
      if (MM_SELECTED_OPTION != MENU_START_ID)
//        TriggerTune(TUNE_MENU_BLIP_A, TUNE_MENU_BLIP_B);
        TriggerSoundEffect(SOUND_MENU_BLIP_ID);
      
      switch (MM_SELECTED_OPTION)
      {
        case MENU_QUADRANT_ID:
          MM_QUADRANT--;
          break;
          
        case MENU_SECTOR_ID:
          if (MM_FIRE_DOWN)
            MM_INITIAL_SECTOR -= 10;
          else
            MM_INITIAL_SECTOR--;
          break;
          
        case MENU_LEVEL_ID:
            MM_LEVEL--;
          break;
          
        case MENU_TV_TYPE_ID: // TV Type: [0]=NTSC, 1=PAL, 2=SECAM
          if (MM_TV_TYPE == 0)
            MM_TV_TYPE = 2;
          else
            MM_TV_TYPE--;
          break;
      }
    }
    else if (!(SWCHA & 0x80) || (MM_FIRE_DOWN))	// joystick right
    {
      if (MM_SELECTED_OPTION != MENU_START_ID)
//        TriggerTune(TUNE_MENU_BLIP_A, TUNE_MENU_BLIP_B);
        TriggerSoundEffect(SOUND_MENU_BLIP_ID);
      
      switch (MM_SELECTED_OPTION)
      {
        case MENU_QUADRANT_ID:
          MM_QUADRANT++;
          break;
          
        case MENU_SECTOR_ID:
          if (!(SWCHA & 0x80) && MM_FIRE_DOWN)
            MM_INITIAL_SECTOR += 10;
          else
            MM_INITIAL_SECTOR++;
          break;
          
        case MENU_LEVEL_ID:
            MM_LEVEL++;
          break;
          
        case MENU_TV_TYPE_ID: // TV Type: [0]=NTSC, 1=PAL, 2=SECAM
          if (MM_TV_TYPE == 2)
            MM_TV_TYPE = 0;
          else
            MM_TV_TYPE++;
          break;
      }
    }
    
    MM_LEVEL &= 3;
    MM_QUADRANT &= 3;
//    MM_INITIAL_SECTOR &= 15;
    
    if ((orig_quadrant  != MM_QUADRANT) ||
        (orig_sector    != MM_INITIAL_SECTOR) ||
        (orig_level     != MM_LEVEL))
    {
      gLastScore = 0;
#ifdef SAVE_HIGH_SCORES
      if (MM_INITIAL_SECTOR == 0)   gHighScore = gHighScores[MM_QUADRANT][MM_LEVEL];
      else                          gHighScore = 0;
#else
      gHighScore = 0;
#endif
    }
  }
}

static void NewGame()
{
  int i;
  MODE = 0;
  gScore = 0;
//  gCondition = 0;
  gReserveLives = 3;
  zil.gInitLevel = 1;
  gSector = MM_INITIAL_SECTOR;
  gSectorBCD = 0;
  gPaused = 0;
  gExtraLife = FIRST_BONUS;
  
  if (gSystemIs7800)
    gColorSwitchState = 0;             // use to debounce 7800 switch for pause detection
  else
    gColorSwitchState = SWCHB & 8;     // we don't care direction, just initial state

  // Convert object colors based on MM_TV_TYPE
  // -4 so the RadarColors also get updated
  for(i=-4;i<OBJECT_COLORS_SIZE;i++)
    gObjectColors[i] = ColorConvert(gObjectColorsNTSC[i]);
  
  // make sure a sample isn't playing so TriggerTune works
  EndSample();
  
  TriggerTune(TUNE_NEW_GAME_A, TUNE_NEW_GAME_B);
  gRoundStart = 1;
}


static void ClearGameDatastreams()
{
  // clears all in-game datastreams
  myMemsetInt((unsigned int*)(DD_BASE + CLEAR_DS_INT), 0, CLEAR_DS_INT_SIZE);
}

const unsigned char energy_offset_x[] =
{
  10, // Horizontal Station Energy Pod 0
  14 // Vertical Station Energy Pod 0
};

const unsigned char energy_offset_y[] =
{
  29, // Horizontal Station Energy Pod 0
  22  // Vertical Station Energy Pod 0
};

static void DisplaySprites()
{
  int i;
  int sprite;
  int shown;
  unsigned char* sprite_shown_ptr;
  unsigned char* sprite_not_shown_ptr;
  int boost;
  
  sprite_not_shown_ptr = gSpriteFlickerList1;
  sprite_shown_ptr = gSpriteFlickerList2;
  
  boost = (MM_TV_TYPE == 2) ? 0 : 2;  // no lumaboost for SECAM
  
  for (i=0;i<MAX_SPRITES;i++)
  {
    sprite = gSpriteFlickerList1[i];
    shown = 0;
    
    if (gSpriteState[sprite])   // non-zero value means sprite is in use
    {
      if (gNonFlickeredSprites == sprite_not_shown_ptr)  boost = 0;
      if (DrawSprite(0, sprite, boost))                  shown = 3; // try to draw using player0
      else if (DrawSprite(1, sprite, boost))             shown = 2; // try to draw using player1
    }
    if (shown)
    {
      *sprite_shown_ptr = sprite;
      sprite_shown_ptr++;
      if (sprite < 8)
      {
        if (gSpriteState[sprite] & SPS_SPRITE_ONSCREEN)
        {
          // if station is open, and core missile is not in place, then display energy field at center
          if ((gStation[sprite].state & (STS_CORE_MISSILE|STS_CORE_OPEN)) == STS_CORE_OPEN)
          {
            gMissileX[MAX_MISSILES] = gSpriteX[sprite] + (energy_offset_x[gSpriteImageID[sprite]] << X_FRACTION_BITS);
            gMissileY[MAX_MISSILES] = gSpriteY[sprite] + (energy_offset_y[gSpriteImageID[sprite]] << Y_FRACTION_BITS);
            
            // 30 fps core animation
            // gMissileState[MAX_MISSILES] = 8 + (gSpriteImageID[sprite]<<2) + ((sprite + (FRAME >> 1)) & 0x03) ;
            
            // 60 fps core animation
            gMissileState[MAX_MISSILES] = 8 + (gSpriteImageID[sprite]<<2) + ((sprite + FRAME) & 0x03) ;
            
            // try to draw with ball first, else opposite player's missile
            if (DrawMissile(4, MAX_MISSILES, 1, gObjectColors[COLOR_STATION_CORE + ((FRAME >> 4) & 0x03)]));
            else DrawMissile(shown, MAX_MISSILES, 1, gObjectColors[COLOR_STATION_CORE + ((FRAME >> 4) & 0x03)]);
          }
        }
      }
    }
    else
    {
      *sprite_not_shown_ptr = sprite;
      sprite_not_shown_ptr++;
    }
  }
  gNonFlickeredSprites = sprite_not_shown_ptr;
  myMemcpy(sprite_not_shown_ptr, gSpriteFlickerList2, sprite_shown_ptr-gSpriteFlickerList2);
}

const unsigned short int game_streams[21] =
{
  P0_DATASTREAM - REPOSITION_SIZE,
  P1_DATASTREAM - REPOSITION_SIZE,
  HMP0_DATASTREAM - REPOSITION_SIZE,
  HMP1_DATASTREAM - REPOSITION_SIZE,
  MISSILE0_DATASTREAM - REPOSITION_SIZE,
  MISSILE1_DATASTREAM - REPOSITION_SIZE,
  BALL_DATASTREAM - REPOSITION_SIZE,
  COLOR_DATASTREAM - REPOSITION_SIZE,
  SIZE_DATASTREAM - REPOSITION_SIZE,
  RADAR_CONTROL,
  RADAR_P0,
  RADAR_P1,
  FORMATION_P0,
  FORMATION_P1,
  PLAYFIELD_DATA + PF0L,
  PLAYFIELD_DATA + PF1L,
  PLAYFIELD_DATA + PF2L,
  PLAYFIELD_DATA + PF0R,
  PLAYFIELD_DATA + PF1R,
  PLAYFIELD_DATA + PF2R,
  OBJECT_COLORS_RAM + COLOR_STATION
};

static void SetGameDatastreams()
{
  int i;
  
  for (i=-REPOSITION_SIZE; i<DATASTREAM_SIZE;i++)
    if (gJumpDataStream[i] == 0)
      gJumpDataStream[i] = NORMAL_KERNEL;
  
  // sets the exit jump to trigger end of drawing the gameplay area
  EXIT_JUMP_DS = EXIT_KERNEL;
  
  for (i=0; i<RADAR_HEIGHT-6;i++)
    gScoreDataStream[i] = SCORE_KERNEL;
  
  EXIT_SCORE_DS = EXIT_SCORE_KERNEL;
  
  for(i=DS_GRP0; i<=DS_STATION_COLOR; i++)
  {
    // surprisingly going direct to pointers and increments didn't save ROM
//    QPTR[i] = game_streams[i] << 20;
//    QINC[i] = 0x100;
    setPointer(i, game_streams[i]);
//dgs    setIncrement(i, 1, 0);
    
  }
//  QINC[DS_STATION_COLOR] = 0;
//  QPTR[33] = (JUMP_DATASTREAM - (REPOSITION_SIZE*2)) << 20;
//dgs  setIncrement(DS_STATION_COLOR,  0, 0);
  setPointer(33,          JUMP_DATASTREAM - (REPOSITION_SIZE*2));
  

  // game area
//  setPointer(DS_GRP0,     P0_DATASTREAM - REPOSITION_SIZE);
//  setPointer(DS_GRP1,     P1_DATASTREAM - REPOSITION_SIZE);
//  setPointer(DS_HMP0,     HMP0_DATASTREAM - REPOSITION_SIZE);
//  setPointer(DS_HMP1,     HMP1_DATASTREAM - REPOSITION_SIZE);
//  setPointer(DS_MISSILE0, MISSILE0_DATASTREAM - REPOSITION_SIZE);
//  setPointer(DS_MISSILE1, MISSILE1_DATASTREAM - REPOSITION_SIZE);
//  setPointer(DS_BALL,     BALL_DATASTREAM - REPOSITION_SIZE);
//  setPointer(DS_COLOR,    COLOR_DATASTREAM - REPOSITION_SIZE);
//  setPointer(DS_SIZE,     SIZE_DATASTREAM - REPOSITION_SIZE);
  
//  setPointer(33,          JUMP_DATASTREAM - (REPOSITION_SIZE*2));
  
//  setIncrement(DS_GRP0,     1, 0);
//  setIncrement(DS_GRP1,     1, 0);
//  setIncrement(DS_HMP0,     1, 0);
//  setIncrement(DS_HMP1,     1, 0);
//  setIncrement(DS_MISSILE0, 1, 0);
//  setIncrement(DS_MISSILE1, 1, 0);
//  setIncrement(DS_BALL,     1, 0);
//  setIncrement(DS_COLOR,    1, 0);
//  setIncrement(DS_SIZE,     1, 0);
  
  //score/radar/lives/formation dislay
//  setPointer(DS_RADAR_CONTROL,  RADAR_CONTROL);
//  setPointer(DS_RADAR_GRP0,     RADAR_P0);
//  setPointer(DS_RADAR_GRP1,     RADAR_P1);
//  setPointer(DS_FORMATION_GRP0, FORMATION_P0);
//  setPointer(DS_FORMATION_GRP1, FORMATION_P1);
//  setPointer(DS_PF0L,           PLAYFIELD_DATA + PF0L);
//  setPointer(DS_PF1L,           PLAYFIELD_DATA + PF1L);
//  setPointer(DS_PF2L,           PLAYFIELD_DATA + PF2L);
//  setPointer(DS_PF0R,           PLAYFIELD_DATA + PF0R);
//  setPointer(DS_PF1R,           PLAYFIELD_DATA + PF1R);
//  setPointer(DS_PF2R,           PLAYFIELD_DATA + PF2R);
//  setPointer(DS_STATION_COLOR,  OBJECT_COLORS_RAM + COLOR_STATION);
  
//  setIncrement(DS_RADAR_CONTROL,  1, 0);
//  setIncrement(DS_RADAR_GRP0,     1, 0);
//  setIncrement(DS_RADAR_GRP1,     1, 0);
//  setIncrement(DS_FORMATION_GRP0, 1, 0);
//  setIncrement(DS_FORMATION_GRP1, 1, 0);
//  setIncrement(DS_PF0L,           1, 0);
//  setIncrement(DS_PF1L,           1, 0);
//  setIncrement(DS_PF2L,           1, 0);
//  setIncrement(DS_PF0R,           1, 0);
//  setIncrement(DS_PF1R,           1, 0);
//  setIncrement(DS_PF2R,           1, 0);
//  setIncrement(DS_STATION_COLOR,  0, 0);
}

const unsigned char character_stream[] =
{
  PF0L,   // 0 - millions digit
  PF1L,   // 1 - hundred thousands digit
  PF1L,   // 2 - ten thousands digit
  PF2L,   // 3 - thousands digit
  PF2L,   // 4 - hundreds digit
  PF0R,   // 5 - tens digit
  PF1R,   // 6 - ones digit
  PF1R,   // 7 - overlaps radar
  PF2R,   // 8 - overlaps radar
  PF2R,   // 9 - after radar
  PF2R    // reserve lives digit, straddles 8 & 9
};

const unsigned char character_offset[] =
{
  4,      // 0 - millions digit
  4,      // 1 - hundred thousands digit
  0,      // 2 - ten thousands digit
  0,      // 3 - thousands digit
  4,      // 4 - hundreds digit
  4,      // 5 - tens digit
  4,      // 6 - ones digit
  0,      // 7 - overlaps radar
  0,      // 8 - overlaps radar
  4,      // 9 - after radar
  2       // reserve lives digit
};

const unsigned char character_flipped[] =
{
  1,      // 0 - millions digit
  0,      // 1 - hundred thousands digit
  0,      // 2 - ten thousands digit
  1,      // 3 - thousands digit
  1,      // 4 - hundreds digit
  1,      // 5 - tens digit
  0,      // 6 - ones digit
  0,      // 7 - overlaps radar
  1,      // 8 - overlaps radar
  1,      // 9 - after radar
  1       // reserve lives digit
};

//const unsigned short int character_flipped = 0x739; // %111 0011 1001

const unsigned char flip_nybble[] =
{
  0,  //  0 0000 -> 0000
  8,  //  1 0001 -> 1000
  4,  //  2 0010 -> 0100
  12, //  3 0011 -> 1100
  2,  //  4 0100 -> 0010
  10, //  5 0101 -> 1010
  6,  //  6 0110 -> 0110
  14, //  7 0111 -> 1110
  1,  //  8 1000 -> 0001
  9,  //  9 1001 -> 1001
  5,  // 10 1010 -> 0101
  13, // 11 1011 -> 1101
  3,  // 12 1100 -> 0011
  11, // 13 1101 -> 1011
  7,  // 14 1110 -> 0111
  15  // 15 1111 -> 1111
};


static void ShowCharacter(int character, int position)
{
  unsigned char temp;
  int i;
  int upper;
  
  upper = (character & 1) ? 0 : 4;    // even character shapes are in upper nybble
  character >>= 1;
  
  for (i=0;i<PF_FONT_HEIGHT;i++)
  {
    temp = PF_FONT[(character * PF_FONT_HEIGHT) + i];
    temp >>= upper;
    temp &= 0x0f;
    if (character_flipped[position]) temp = flip_nybble[temp];
//    if (character_flipped & (1 << position))    temp = flip_nybble[temp];
    temp <<= character_offset[position];
    gPlayfield[character_stream[position] + i + 1] |= temp;
  }
}

// only the 3 pixel wide letters
#define LAST_CHARACTER 42

// includes wide M N and W
//#define LAST_CHARACTER 48

const signed char formation_leader_offset[] =
{
  0, // line
  3, // arrow
  6, // square
  6  // diamond
};

const signed char formation_y_offset[] =
{
  3, // line
  6, // arrow
  3, // square
  3  // diamond
};

const unsigned char paused[6] = {21, 10, 20, 16, 14, 13};
//const unsigned char sector[6] = {16, 14, 12, 17, 0, 18};  // used 20 more bytes of ROM

static void PrepScoreRadar()
{
  int i;
  int x;
  int y;
  int score;

  // set condition color
  if (zil.gConditionRed)                      i = 2;
  else if (gPriorFrameEnemyMissileOnscreen)   i = 1;
  else                                        i = 0;
  x = gObjectColors[COLOR_CONDITION_GREEN + i];
  // flash condition red
  if ((i == 2) && (FRAME & 0x10))             x = 0;
  gRadarControl[0] = x;
  // end set condition color
  
  // set score color
  gRadarControl[3] = ColorConvert(SCORE_COLOR);
  
  if (gPaused)
  {
    for(i=0;i<6;i++)
      ShowCharacter(paused[i],i+2);
    if (FRAME & 64)
      gRadarControl[3] = 0;
    gRadarControl[0] = 0;
    return;
  }

  if (SWCHB & 0x40)
  { // left difficulty = A
    
    // time remaining in VB and OS
    ShowCharacter(VB_TIME >> 4, 1);
    ShowCharacter(VB_TIME & 15, 2);

    ShowCharacter(OS_TIME >> 4, 5);
    ShowCharacter(OS_TIME & 15, 6);
  }
  else if (gSectorBCD == 0)
  {
    // prevent SECTOR<blank> from showing in score area when starting a new game
    // as SECTOR will shift left on the next frame
    return;
  }
  else if (gRoundStart)
  {
    // show SECTOR ###
    i=9;
    score = gSectorBCD;
    while(score)
    {
      ShowCharacter(score & 0x0f, i);
      i--;
      score >>= 4;
    }
    ShowCharacter(18, --i); // R
    ShowCharacter( 0, --i); // O
    ShowCharacter(17, --i); // T
    ShowCharacter(12, --i); // C
    ShowCharacter(14, --i); // E
    ShowCharacter(16, --i); // S
    
    return;
  }
  else
  {
    // show score
    score = gScore;
    
    for (i=6;i>=0;i--)
    {
      ShowCharacter(score & 0x0f, i);
      score >>= 4;
      if (score == 0) break;
    }
  }
  
  // prep radar frame
  gRadarP0[ 0] = 0xc0;  // top left
  gRadarP0[ 1] = 0x80;  // top left
  gRadarP1[ 0] = 0x03;  // top right
  gRadarP1[ 1] = 0x01;  // top right
  gRadarP0[18] = 0x80;  // bottom left
  gRadarP0[19] = 0xc0;  // bottom left
  gRadarP1[18] = 0x01;  // bottom right
  gRadarP1[19] = 0x03;  // bottom right
  
  // show reserve lives, or X if over 9
  i = (gReserveLives <= 9) ? gReserveLives : 19 ;// 19 = X
  ShowCharacter(i, 10);
  
  if (gJumpDataStream[0] != INIT_LEVEL_KERNEL)
  {
    // update station location on radar display
    for (i=0;i<8;i++)
    {
      if (gSpriteState[i])
      {
        x = gSpriteX[i] >> (X_FRACTION_BITS + RADAR_SHIFT);
        y = gSpriteY[i] >> (Y_FRACTION_BITS + RADAR_SHIFT);
        gRadarP1[y+2] |= 1 << (7-x);
      }
    }
    
    // update formation location on radar display
    if (zil.gFormationAttackPattern)
    {
      x = gSpriteX[FORMATION_SPRITE] >> (X_FRACTION_BITS + RADAR_SHIFT);
      y = gSpriteY[FORMATION_SPRITE] >> (Y_FRACTION_BITS + RADAR_SHIFT);
      gRadarP0[y+2] |= 1 << (7-x);
      
      // update formation leader location on formation display
      i = formation_leader_offset[zil.gFormationAttackPattern-1];
      gFormationP0[i++] |= 8;
      gFormationP0[i] |= 8;
      
      // update formation followers on radar display
      myMemcpy((unsigned char *)DD_BASE + FORMATION_P1 + formation_y_offset[zil.gFormationAttackPattern - 1],
               (unsigned char *)((unsigned int)gImageGraphics[FORMATION_ID + zil.gFormationAttackPattern - 1]),
               gImageHeight[FORMATION_ID + zil.gFormationAttackPattern - 1]);
    }
    
    // update player location on radar display
    if (gSpriteState[PLAYER_SPRITE])
    {
      if ((FRAME & 8) == 8)
      {
        x = -(gSpriteX[PLAYER_SPRITE] >> (X_FRACTION_BITS + RADAR_SHIFT));
        x <<= 4;                // change to HMOVE value
        QUEUE[RADAR_CONTROL + 1] = x;   // adjust ball position
        
        y = gSpriteY[PLAYER_SPRITE] >> (Y_FRACTION_BITS + RADAR_SHIFT);
        gPlayfield[y + PF0L] |= 2;
      }
    }
  }
}

const unsigned short int reposition_kernel[5][11] =
{
  {
    RESP0STROBE23, RESP0STROBE28,
    RESP0STROBE33, RESP0STROBE38,
    RESP0STROBE43, RESP0STROBE48,
    RESP0STROBE53, RESP0STROBE58,
    RESP0STROBE63, RESP0STROBE68,
    RESP0STROBE73
  },
  {
    RESP1STROBE23, RESP1STROBE28,
    RESP1STROBE33, RESP1STROBE38,
    RESP1STROBE43, RESP1STROBE48,
    RESP1STROBE53, RESP1STROBE58,
    RESP1STROBE63, RESP1STROBE68,
    RESP1STROBE73
  },
  {
    RESM0STROBE23, RESM0STROBE28,
    RESM0STROBE33, RESM0STROBE38,
    RESM0STROBE43, RESM0STROBE48,
    RESM0STROBE53, RESM0STROBE58,
    RESM0STROBE63, RESM0STROBE68,
    RESM0STROBE73
  },
  {
    RESM1STROBE23, RESM1STROBE28,
    RESM1STROBE33, RESM1STROBE38,
    RESM1STROBE43, RESM1STROBE48,
    RESM1STROBE53, RESM1STROBE58,
    RESM1STROBE63, RESM1STROBE68,
    RESM1STROBE73
  },
  {
    RESBLSTROBE23, RESBLSTROBE28,
    RESBLSTROBE33, RESBLSTROBE38,
    RESBLSTROBE43, RESBLSTROBE48,
    RESBLSTROBE53, RESBLSTROBE58,
    RESBLSTROBE63, RESBLSTROBE68,
    RESBLSTROBE73
  }
};

static int DrawSprite(int stream, int index, int lumaboost)
  {
    // if possible, copies image to specified datastream
    // returns 1 if successful, 0 if failed.
    static int y;
    static int x;
    static int height;
    static int imageID;
    static int offset;  // offset from top of sprite, so image can smoothly move off top
    static int size;
    static int width;
    
    int resp;
    int position;
    int i;
    unsigned char * playerstream;
    unsigned char * hmpstream;
    int mask;
    
    // init only needed when called for stream 0 as same values would have been derived for stream 1
    if (stream == 0)
    {
      gSpriteState[index] &= ~SPS_SPRITE_ONSCREEN;
      if (gSpriteState[index] & SPS_4X_SPRITE)
      {
        size = 7;
        width = 32;
      }
      else if (gSpriteState[index] & SPS_2X_SPRITE)
      {
        size = 5;
        width = 16;
      }
      else
      {
        size = 0;
        width = 8;  // no masking needed
      }
      
      x = ((gSpriteX[index] >> X_FRACTION_BITS) - (gWindowX >> X_FRACTION_BITS)) & (WORLD_WIDTH-1);
      y = ((gSpriteY[index] >> Y_FRACTION_BITS) - (gWindowY >> Y_FRACTION_BITS)) & (WORLD_HEIGHT-1);
      
      if (x > (WORLD_WIDTH / 2))  x -= WORLD_WIDTH;
      if (y > (WORLD_HEIGHT / 2)) y -= WORLD_HEIGHT;
      
      if (x+width < 8) return 1;    // sprite is offscreen, no need to draw so treat like we did
      if (x > 159) return 1;        // sprite is offscreen, no need to draw so treat like we did
      
      imageID = gSpriteImageID[index];
      height = gImageHeight[imageID];
      
      if (y < 0)
      {
        if (-y >= height)
          return 1;   // sprite is offscreen, no need to draw so treat like we did
        offset = -y;
        height -= offset;
        y = 0;
      }
      else if (y >= DATASTREAM_SIZE)
        return 1;           // sprite is offscreen, no need to draw so treat like we did
      else
        offset = 0;
      
      if (y+height >= DATASTREAM_SIZE)
        height = DATASTREAM_SIZE - y; // to smoothly scroll image off bottom
      
//  return 0; // uncomment to test with only player 1
      
      gSpriteState[index] |= SPS_SPRITE_ONSCREEN;
      
      // set stream 0 specific values
      playerstream    = gP0DataStream;
      hmpstream       = gHMP0DataStream;
    }
    else
    {
      // set stream 1 specific values
      playerstream    = gP1DataStream;
      hmpstream       = gHMP1DataStream;
    }
    
    if      (gJumpDataStream[y-1] == 0) position = -1;  // no conflict
    else if (gJumpDataStream[y-2] == 0) position = -2;  // conflict with other player
    else if (gJumpDataStream[y-3] == 0) position = -3;  //            or ball (station core)
    else if (gJumpDataStream[y-4] == 0) position = -4;  //            or missile (station core)
    else                                return 0;       //            or self
    
    // check if the player is already in use for the scanlines we need
    for (i = (y + position); i <= (y + height + 1) && i <= DATASTREAM_SIZE;i++)
    {
      if (hmpstream[i] != 0)  // bit 3 is set for active as player graphics could have gaps in them
        return 0;             // abort, player in use
    }
    
    // sprite can go here, so copy it in!
    if (index < 8)
    {
      // copy over station images which are in RAM so they can have pods removed
      myMemcpy(playerstream + y,
               (unsigned char *)DD_BASE + STATION_IMAGES + index*STATION_HEIGHT + offset,
               height);
    }
    else
    {
      // all other images are in ROM
      myMemcpy(playerstream + y,
               (unsigned char *)offset + gImageGraphics[imageID],
               height);
    }
    
    // check for masking on 2x and 4x sprites
    if (width != 8)
    {
      mask = 0xff;
      if (x < 0)                  // mask on left edge
      {
        if (width == 32)
          mask >>= ((6-x) >> 2);
        else
          mask >>= ((6-x) >> 1);
      }
      else if (x + width > 160)   // check if mask needed on right edge
      {
        if (width == 32)
          mask <<= ((x-128) >> 2);
        else
          mask <<= ((x-144) >> 1);
      }
      if (mask != 0xff)
      {
        resp = 0;
        // apply the mask to the player stream.  Sum the results in resp
        for (i = (y + position); i < (y + height) && i < DATASTREAM_SIZE;i++)
          resp += (playerstream[i] &= mask);
        // if mask removes the image, then don't continue processing this sprite.
        // this happens with the Horizontal Station when it gets masked on the
        // left side of the screen.
        if (resp == 0)
          return 1;
      }
    }
    
    // "wiggle" large sprites so they look like they're hi-res
    if (imageID <= SHIFTING_SPRITES)
    {
      myMemcpy(hmpstream + y,
               (unsigned char *)offset + gImageShift[imageID],
               height);
      x += hmpstream[y] & 0x07;               // adjust x based on first row of "wiggled" image
      hmpstream[y] &= 0x0f;
      hmpstream[y + height] = 0x08;           // prevent early reuse
    }
    else
    {
      // set bit 3 in the HMOVE datastream for in-use check
      myMemset(hmpstream + y + position, 8, height + 1 - position);
    }
    
    // 2x and 4x objects get displayed 1 pixel to the right, so compensate
    if (size) x--;
    
    x++;    // reposition routines are off by 1 pixel from 6507 PosObject
    
    if (x <= 4)     // HMOVE with cycle 23 RESPx has issues,
      x += 160;   // this compensates
    
    resp = 0;
    while (x > 15)
    {
      x -= 15;
      resp++;
    }
    if (resp > 10)		resp = 0;	// prevent a crash
    
    if ((resp == 0) && (x >= 5))   // HMOVE with cycle 23 RESPx has issues,
      x -= 5;                     // this compensates
    
    x ^= 7;
    x <<= 4;
    
    gJumpDataStream[y + position]           = reposition_kernel[stream][resp]; // specify which kernel to use to strobe RESPx on correct cycle
    //  hmpstream[y + position]               = 8;                // prevent premature sprite reuse - next line used to be +1
    hmpstream[y + position]                 = x | 8;            // set fine-positioning, prevent premature sprite reuse
    gSizeDataStream[y + position]           = size;
    gColorDataStream[y + position]          = gSpriteColor[index] + lumaboost;
    
    return 1;
  }

const unsigned char ship_x_adj[] = { 0, 1, 1, 1, 0, 0, 0, 0 };
//const unsigned char ship_x_adj = 0x70; // %0111 0000

const unsigned char joystick_direction[] = { 9, 9, 9, 9, 9, 3, 1, 2, 9, 5, 7, 6, 9, 4, 0, 9};
//  7 0 1
//  6   2
//  5 4 3

const unsigned char shot_shape[] = { SHOT_VERTICAL_ID, SHOT_DIAGONAL_TRBL_ID, SHOT_HORIZONTAL_ID, SHOT_DIAGONAL_TLBR_ID};

const signed char shot_offset_x[] = { 4, 8, 9,  8,  4,  -2, -4, -2};

const signed char shot_offset_y[] = {-5, -2, 7, 15, 16, 15, 7, -3};

static void ProcessJoystick()
{
  int i;
  int x;
  int y;
  int direction;
  int free_shot[2];
  int offset;
  
  x = 0;
  y = 0;
  
  if (zil.gRoundCleared)
    return;
  
  if (gRoundStart)
  {
    // gRoundStart
    // 0 = round fully active
    // 1 = wait for end of opening tune then Say "Blast Off" (show Score/Radar/Lives)
    // 2 = wait for end of "Blast Off" then start ship moving
    
    if (gRoundStart == 1) // wait for end of opening tune then Say "Blast Off" (show Score/Radar/Lives)
    {
      if (SOUND_VOL[gActiveSound[0]] == 0 && SOUND_VOL[gActiveSound[1]] == 0) // check that tune has finished
      {
        gRoundStart = 2;
        TriggerSample(SAMPLE_BLAST_OFF);
      }
    }
    else if (gRoundStart == 2)
    {
      if (gCurrentSample == 0)
        gRoundStart = 0;
    }
    gIdleConditionRedTimer = gIdleConditionRedMax;  // reset idle timer
    return;
  }
  
  direction = joystick_direction[SWCHA >> 4];
  
  if (gSpriteState[PLAYER_SPRITE] & SPS_DEADLY_TO_ENEMY)  // true if player is still alive
  {
    if (direction < 8)
      zil.gShipDir = direction;
    gSpriteImageID[PLAYER_SPRITE]   = SHIP_ID + zil.gShipDir;
  }
  else
  {
    // player's exploding - disable firebutton
    INPT4 = 0xff;
  }
  
#ifdef RIGHT_DIFFICULTY_FREEZE
  if (SWCHB & 0x80)   // Right Difficulty A = stationary ship
    return;
#endif
  
  x = speed_dir_x[zil.gShipDir * 4];
  y = speed_dir_y[zil.gShipDir * 4];
  if (zil.gPlayerSpeed != 128)
    zil.gPlayerSpeed++;
  
  x = (zil.gPlayerSpeed * x) / 128;
  y = (zil.gPlayerSpeed * y) / 128;
  
  // move window based on joystick input
  gWindowX += x;
  gWindowY += y;
  
  // recenter player
  gSpriteX[PLAYER_SPRITE] = gWindowX + ((80 + ship_x_adj[(zil.gShipDir & 7)]) << X_FRACTION_BITS);
//  gSpriteX[PLAYER_SPRITE] = gWindowX + ((80 + ((ship_x_adj >> (zil.gShipDir & 7)) & 1)) << X_FRACTION_BITS);
  gSpriteY[PLAYER_SPRITE] = gWindowY + ((((DATASTREAM_SIZE - SHIP_HEIGHT) /2)) << Y_FRACTION_BITS);

  if (gSpriteState[PLAYER_SPRITE] & (SPS_EXPLOSION | SPS_GAMEOVER))
  {
    gSpriteX[PLAYER_SPRITE] -= 4 << X_FRACTION_BITS;
    gSpriteY[PLAYER_SPRITE] -= 8 << Y_FRACTION_BITS;
  }
  
  // player shots don't shift with the window, so compenstate their X Y values
  for (i=0;i<4;i++)
  {
    if (gMissileState[i])
    {
      gMissileX[i] += x;
      gMissileY[i] += y;
    }
  }
  
  // update starfield based on player's movement
  zil.gStarfieldXoffset -= (x >> STARFIELD_FRAC_BITS);
  gStarfieldYoffset -= (y >> STARFIELD_FRAC_BITS);
  if (y > 0)
  {
    if (gStarfieldYoffset < 0)
    {
      gStarfieldYoffset += 2 << Y_FRACTION_BITS;
      zil.gStarfieldTopStar++;
      zil.gStarfieldTopStar &= 127;
    }
  }
  else
  {
    if (gStarfieldYoffset >> Y_FRACTION_BITS == 2)
    {
      gStarfieldYoffset -= 2 << Y_FRACTION_BITS;
      zil.gStarfieldTopStar--;
      zil.gStarfieldTopStar &= 127;
    }
  }
  
  // gun processing
  if (zil.gGunCoolDown)
    zil.gGunCoolDown--;
  else if (!(INPT4 & 0x80))  // fire down
  {
    x = 0;
    for (i=0;i<4 && x < 2;i++)
    {
      if (gMissileState[i] == 0)
      {
        free_shot[x] = i;
        x++;
      }
    }
    
    if (x == 2 ) // && zil.gGunCoolDown == 0)
    {
      TriggerSoundEffect(SOUND_PLAYER_SHOOTS_ID);
      zil.gGunCoolDown = GUN_COOL_DOWN;
      for (i=0;i<2;i++)
      {
        x = free_shot[i];
        gMissileState[x] = 0x80 + shot_shape[zil.gShipDir & 3];
        gMissileDirSpeed[x] = (zil.gShipDir + i*4) & 7;
        offset =((zil.gShipDir + i * 4) & 7);
        gMissileX[x] = gSpriteX[PLAYER_SPRITE] + ((shot_offset_x[offset] - ship_x_adj[zil.gShipDir]) << X_FRACTION_BITS);
//        gMissileX[x] = gSpriteX[PLAYER_SPRITE] + ((shot_offset_x[offset] - ((ship_x_adj >> zil.gShipDir) & 1)) << X_FRACTION_BITS);
        gMissileY[x] = gSpriteY[PLAYER_SPRITE] + (shot_offset_y[offset] << Y_FRACTION_BITS);
      }
    }
  }
}

static int DrawMissile(int stream, int index, int firstcall, int color)
{
  // if possible, preps missile display in the specified datastream
  // returns 1 if drawn, 0 if failed, 2 if off-screen so shots can be turned off.
  // streams are 2=missile0, 3=missile1, 4=ball
  static int x;
  static int y;
  static int height;
  static int offset;  // offset from top of missile, so image can smoothly move off top
  static int imageID;
  static unsigned short int *image;
  
  int resp;
  int position;
  int i;
  
  unsigned char * objectstream; // HMxx in upper nybble, ENAxx in lower nybble
  
  // init only needed for firstcall
  if (firstcall)
  {
    x       = ((gMissileX[index] >> X_FRACTION_BITS) - (gWindowX >> X_FRACTION_BITS)) & (WORLD_WIDTH-1);
    y       = ((gMissileY[index] >> Y_FRACTION_BITS) - (gWindowY >> Y_FRACTION_BITS)) & (WORLD_HEIGHT-1);
    
    if (x > (WORLD_WIDTH / 2)) x -= WORLD_WIDTH;
    if (y > (WORLD_HEIGHT / 2)) y -= WORLD_HEIGHT;
    
    if (x <= 0)   return 2; // missile is offscreen
    if (x > 159)  return 2; // missile is offscreen
    
    imageID = gMissileState[index] & 0x0f;
    height  = gImageHeight[imageID + FIRST_MISSILE_GRAPHIC_ID];
    image   = (unsigned short int *)(0+gMissileGraphics[imageID]);
    
    if (y < 0)
    {
      if (-y >= height)
        return 2;   // missile is offscreen
      offset = -y;
      height -= offset;
      y = 0;
    }
    else if (y >= DATASTREAM_SIZE)
      return 2;     // missile is offscreen
    else
      offset = 0;
    
    if (y+height >= DATASTREAM_SIZE)
      height = DATASTREAM_SIZE - y;   // to smoothly scroll image off bottom
  }
  
  if (stream == 2)
    objectstream = gMissile0DataStream;
  else if (stream == 3)
    objectstream = gMissile1DataStream;
  else // stream == 4
    objectstream = gBallDataStream;
  
  if      (gJumpDataStream[y-0] == 0) position =  0;  // no conflict
  else if (gJumpDataStream[y-1] == 0) position = -1;  // conflict with player
  else if (gJumpDataStream[y-2] == 0) position = -2;  //            or other player
  else if (gJumpDataStream[y-3] == 0) position = -3;  //            or ball
  else if (gJumpDataStream[y-4] == 0) position = -4;  //            or other missile
  else                                return 0;       //            or self
  
  // check if the object is already in use for the scanlines we need
  for (i = (y + position); i <= (y + height + 1) && i <= DATASTREAM_SIZE;i++)
  {
    if (objectstream[i] != 0)
      return 0;
  }
  
  // missile can go here, so copy it in!
  for(i=0; i<height;i++)
    objectstream[i + y]       = image[i + offset];
  
  objectstream[y+height] = 0x08;  // prevent early reuse
  
  if (position) // position < 0
  {
    for(i=position; i<0;i++)
      objectstream[y+i] = 0x08;   // prevent reuse
  }
//  else // position == 0
//  {
//    gScore = ((image[offset] >> 8) & 0x0f);
//    x += ((image[offset] >> 8) & 0x0f); // lower nybble of upper byte of image is X offset
//  }

  x += (image[offset] >> 8) & 0x0f;
  objectstream[y] &= 0x0f;
    
  // missiles' X is off by 1, and reposition routines are off by 1, so add 2
  x += 2;
  
  if (x <= 4)     // HMOVE with cycle 23 RESPx has issues,
    x += 160;   // this compensates
  
  resp = 0;
  while (x > 15)
  {
    x -= 15;
    resp++;
  }
  if (resp > 10)		resp = 0;	// prevent a crash
  
  if ((resp == 0) && (x >= 5))   // HMOVE with cycle 23 RESP1 has issues,
    x -= 5;                     // this compensates
  
  x ^= 7;
  x <<= 4;
  
  gJumpDataStream[y + position]   = reposition_kernel[stream][resp]; // set which cycle to strobe RESPx
  objectstream[y + position]     &= 0x0f;
  objectstream[y + position]     |= x;
  gSizeDataStream[y + position]   = (image[offset] >> 8) & 0xf0; // upper nybble of upper byte of image is size
  gColorDataStream[y + position]  = color;  // only used when drawing ball
  
  return 1;
}


static void DisplayMissiles()
{
  int i;
  int missile;
  int shown;
  int ds0;
  int ds1;
  unsigned char* missile_shown_ptr;
  unsigned char* missile_not_shown_ptr;
  int offscreen;
  
  //  return;
  
  missile_not_shown_ptr = gMissileFlickerList1;
  missile_shown_ptr = gMissileFlickerList2;
  
  // alternate datastreams to blend color of the missiles, helping them to
  // stand out more
  ds0 = FRAME & 1;
  ds1 = 1 - ds0;
  ds0 += 2;
  ds1 += 2;
  
  for (i=0;i<MAX_MISSILES;i++)
  {
    missile = gMissileFlickerList1[i];
    shown = 0;
    
    if (gMissileState[missile])   // non-zero value means missile is in use
    {
      if ((offscreen = DrawMissile(ds0, missile, 1, 15)) != 0)       shown = 1;  // try to draw using a missile
      else if ((offscreen = DrawMissile(ds1, missile, 0, 15)) != 0)  shown = 1;  // try to draw using other missile
      else if ((offscreen = DrawMissile(4, missile, 0, 15)) != 0)    shown = 1;  // try to draw using ball
    }
    if (shown)
    {
      *missile_shown_ptr = missile;
      missile_shown_ptr++;
      if (offscreen == 2)
        gMissileState[missile] = 0; // if missile moved fully offscreen turn it off
    }
    else
    {
      *missile_not_shown_ptr = missile;
      missile_not_shown_ptr++;
    }
  }
  gNonFlickeredMissiles = missile_not_shown_ptr;
  myMemcpy(missile_not_shown_ptr, gMissileFlickerList2, missile_shown_ptr-gMissileFlickerList2);
}


static void FixSizes()
{
  int i;
  unsigned char player[2];
  unsigned char missile[2];
  
  player[0]   = 0;
  player[1]   = 0;
  missile[0]  = 0;
  missile[1]  = 0;
  
  for(i=-REPOSITION_SIZE;i<DATASTREAM_SIZE-1;i++)
  {
    // look for Jumps for player or missile repositioning
    if (gJumpDataStream[i] >= FIRST_RESP0) // && gJumpDataStream[i] < FIRST_RESBL)
    {
      if (gJumpDataStream[i] < FIRST_RESM0) // are they player jumps?
      {
        if (gJumpDataStream[i] < FIRST_RESP1) // are they player 0 jumps?
        {
          player[0] = gSizeDataStream[i];    // get new player size
          gSizeDataStream[i] |= missile[0];  // merge in current missile size
        }
        else
        {
          player[1] = gSizeDataStream[i];    // get new player size
          gSizeDataStream[i] |= missile[1];  // merge in current missile size
        }
      }
      else                                    // it's a missile jump
      {
        if (gJumpDataStream[i] < FIRST_RESM1) // are they missile 0 jumps?
        {
          missile[0] = gSizeDataStream[i];   // get new missile size
          gSizeDataStream[i] |= player[0];   // merge in current player size
        }
        else
        {
          missile[1] = gSizeDataStream[i];   // get new missile size
          gSizeDataStream[i] |= player[1];   // merge in current player size
        }
      }
    }
  }
}

static void DisplayStars()
{
  int i;
  int x;
  int y;
  int resp;
  int blink;
  int x_offset;
  
  y = gStarfieldYoffset >> Y_FRACTION_BITS;
  blink = ((FRAME >> 5) + zil.gStarfieldTopStar) & 7;
  i = zil.gStarfieldTopStar;
  x_offset = zil.gStarfieldXoffset >> X_FRACTION_BITS;
  
  while (y < DATASTREAM_SIZE - 1)
  {
    if (blink && (gJumpDataStream[y]  == 0) && gBallDataStream[y] == 0 && gBallDataStream[y+1] == 0)
    {
      x = (gStarfieldX[i] + x_offset) & 255;
      
      if (x > 7 && x < 160)
      {
        resp = 0;
        while (x > 15)
        {
          x -= 15;
          resp++;
        }
        if (resp > 10)		resp = 0;	// prevent a crash
        
        if ((resp == 0) && (x >= 5))   // HMOVE with cycle 23 RESP1 has issues,
          x -= 5;                     // this compensates
        
        x ^= 7;
        x <<= 4;
        
        gJumpDataStream[y]          = reposition_kernel[4][resp]; // set which cycle to strobe RESPx
        gBallDataStream[y]          = x | 2;
        gColorDataStream[y]         = gStarfieldColor[i];
// should already be 0        gSizeDataStream[y]          = 0; //i << 4;
      }
    }
    y += 2;
    blink++;
    blink &= 7;
    
    i++;
    i &= 127;
  }
}


// formation spacing
#define FX 9    // originally 12
#define FY 16   // originally 18
const signed char formation_x[4][5] =
{
  {   0*FX,   0*FX,  0*FX,  0*FX,   0*FX }, // line
  {   0*FX,  -1*FX,  1*FX, -2*FX,   2*FX }, // arrow
  {   0*FX,  -1*FX,  1*FX, -1*FX,   1*FX }, // X
  {   0*FX,   0*FX, -1*FX,  1*FX,   0*FX }  // diamond
  
//  {  0,   0,   0,   0,   0}, // line
//  {  0, -12,  12, -24,  24}, // arrow
//  {  0, -12,  12, -12,  12}, // X
//  {  0,   0, -12,  12,   0}  // diamond
};

const signed char formation_y[4][5] =
{
  {  -2*FY,  -1*FY,   0*FY,   1*FY,   2*FY}, // line
  {  -1*FY,   0*FY,   0*FY,   1*FY,   1*FY}, // arrow
  {   0*FY,  -1*FY,  -1*FY,   1*FY,   1*FY}, // X
  {   0*FY,  -1*FY,   0*FY,   0*FY,   1*FY}  // diamond
//  {  0,  18,  36,  54,  72}, // line
//  {  0,  18,  18,  36,  36}, // arrow
//  {  0, -18, -18,  18,  18}, // X
//  {  0, -18,   0,   0,  18}  // diamond
};

const unsigned char formation_ship[8] =
{
              // Sector AND
  1,          //    1    0    only I-Type formations
  1, 2, 2,    //    2    3    50% I-Type, 50% P-Type
  1, 2, 0, 0  //    3+   7    37.5% I-Type, 37.5% P-Type, 25% E-Type
};

static void MoveFormation()
{
  int i;
  int s;
  int t;
  int dir;
  int count;
  
  if (zil.gFormationAttackPattern || zil.gFormationShipCount)
  {
    // change direction every 8th frame
    if ((FRAME & 7) == 0)
    {

      if (zil.gFormationAttackPattern)
      {
        // leader intact, move towards player
        dir = gSpriteState[FORMATION_SPRITE] & SPS_DIRECTION;
        dir = DirectionToPlayer(gSpriteX[FORMATION_SPRITE],
                                gSpriteY[FORMATION_SPRITE],
                                dir,
                                DTP_SPRITE,
                                DTP_TURN);
        for(i=FORMATION_SPRITE;i<SPY_SPRITE;i++)
        {
          if (gSpriteState[i] & SPS_DEADLY_TO_PLAYER)
          {
            gSpriteState[i] &= ~SPS_DIRECTION;
            gSpriteState[i] |= dir;
            gSpriteImageID[i] = gFormationShipType + (dir >> 1);
          }
        }
      }
      else
      {
        // leader destroyed, disperse
        for(i=FORMATION_SPRITE+1;i<SPY_SPRITE;i++)
        {
          if (gSpriteState[i] & SPS_DEADLY_TO_PLAYER)
          {
            if (gSpriteState[i] & SPS_SPRITE_ONSCREEN)
            {
              dir = TurnAwayFromPlayer(gSpriteX[i],
                                       gSpriteY[i],
                                       /*DTP_SPRITE, */
                                       gSpriteState[i] & SPS_DIRECTION);
              
              gSpriteState[i] &= ~SPS_DIRECTION;
              gSpriteState[i] |= dir;
              gSpriteImageID[i] = gFormationShipType + (dir >> 1);
            }
            else
            {
              // ship got away
              gSpriteState[i] = 0;
            }
          }
        }
      }
    }
    
    // move ships
    count = 0;
    for(i=FORMATION_SPRITE;i<SPY_SPRITE;i++)
    {
      if (gSpriteState[i] & SPS_SPRITE_ONSCREEN)
        gPriorFrameEnemyMissileOnscreen = 1;  // used to show condition yellow

      if (gSpriteState[i] & SPS_DEADLY_TO_PLAYER)
      {
        count++;
        dir = gSpriteState[i] & SPS_DIRECTION;
        gSpriteX[i] += (gFormationSpeed * speed_dir_x[dir]) / 128;
        gSpriteY[i] += (gFormationSpeed * speed_dir_y[dir]) / 128;
      }
    }
    if (count == 0)
      zil.gFormationAttackPattern = zil.gFormationShipCount = 0;
  }
  else
  {
    // spawn a formation
   
    // check to make sure formation sprites are not in use, they could still be
    // showing explosions or bonus points
    count = 0;
    for (i = FORMATION_SPRITE; i< FORMATION_SPRITE + 5; i++)
      if (gSpriteState[i])
        count++;
    
    // work is spread across frames due to performance reasons
    // FRAME & 3 FEATURE
    // --------- ------------
    //     0     spawn I-Type missile
    //     1     spawn P-Type missile
    //     2     spawn Formation
    //     3     spawn Spy Ship
    
    // only check 1 station per spawn attempt
    s = (FRAME >> 2) & 7;
    
    if (gFormationTimer)
      gFormationTimer--;
    else if (   (FRAME & 3) == 2                                      // spawn formation frame? (see comment above)
             && (count == 0)                                          // sprites are free to use?
             && (gSpriteState[s] & SPS_DEADLY_TO_PLAYER)              // station is active?
             && ((gSpriteState[s] & SPS_SPRITE_ONSCREEN) == 0)        // launch station is offscreen?
             && gCurrentSample == 0                                   // no sample in play
             && (gSpriteState[PLAYER_SPRITE] & SPS_DEADLY_TO_ENEMY))  // player is alive?
    {
      dir = DirectionToPlayer(gSpriteX[s],
                              gSpriteY[s],
                              0,
                              DTP_SPRITE,
                              DTP_DIRECTION);
      zil.gFormationShipCount = 5;
      zil.gFormationAttackPattern = (getRandom32() & 3) + 1;
      t = formation_ship[getRandom32() & gFormationMask];
      gFormationShipType = (t * 16) + ETYPE_ID;
      for (i = FORMATION_SPRITE; i< FORMATION_SPRITE + 5; i++)
      {
        gSpriteState[i]     = SPS_DEADLY_TO_PLAYER | SPS_ENEMYSHIP | dir;
        gSpriteX[i]         = gSpriteX[s] + (formation_x[zil.gFormationAttackPattern-1][i-FORMATION_SPRITE] << X_FRACTION_BITS);
        gSpriteY[i]         = gSpriteY[s] + (formation_y[zil.gFormationAttackPattern-1][i-FORMATION_SPRITE] << Y_FRACTION_BITS);
        gSpriteImageID[i]   = gFormationShipType + (dir >> 2);
        gSpriteColor[i]     = (i == FORMATION_SPRITE)
        ? gObjectColors[COLOR_E_LEAD + t]
        : gObjectColors[COLOR_E + t];
      }
      TriggerSample(SAMPLE_BATTLESTATIONS);
      
      gFormationTimer = ResetTimer(gFormationMinTime, gFormationRangeTime);
      
      // count for when Spy Ship can launch
      if (gFormationSpyCountdown)
        gFormationSpyCountdown--;
    }
  }
}

static void MoveSpyShip()
{
  int dir;
//  int x;
//  int y;
  
  // spy ship only appears after a certain number of formations have appeared
  if (gFormationSpyCountdown) return;
  
  if (gSpriteState[SPY_SPRITE])
  {
    if (!(gSpriteState[SPY_SPRITE] & SPS_SPRITE_ONSCREEN))
    {
      if (gSpriteImageID[SPY_SPRITE] >= SPY_ID && gSpriteImageID[SPY_SPRITE] < SPY_ID + 16)
      {
        // spy ship made it across the screen
      
        // any more passes for this spy ship?
        if (gSpyPasses)
        {
          // turn around for another pass
          gSpyPasses--;
          dir = gSpriteState[SPY_SPRITE] & SPS_DIRECTION;
          dir += 16;
          dir &= 31;
          dir = DirectionToPlayer(gSpriteX[SPY_SPRITE],
                                  gSpriteY[SPY_SPRITE],
                                  dir,
                                  DTP_SPRITE,
                                  DTP_DIRECTION);
          gSpriteState[SPY_SPRITE] &= ~SPS_DIRECTION;
          gSpriteState[SPY_SPRITE] |= dir;
          
          // do an extra move to make sure it ends up back onscreen
          gSpriteX[SPY_SPRITE] += (gSpySpeed * speed_dir_x[dir]) / 32;  // was 64
          gSpriteY[SPY_SPRITE] += (gSpySpeed * speed_dir_y[dir]) / 32;  // was 64
        }
        else
        {
          // spy ship got away, trigger condition red
          TriggerSample(SAMPLE_CONDITION_RED);
          zil.gConditionRed = 1;
          gSpriteState[SPY_SPRITE] = 0;
        }
      }
      else // image not spy ship, so it was destroyed
        gSpriteState[SPY_SPRITE] = 0;

    }
    
    if (gSpriteState[SPY_SPRITE] & SPS_DEADLY_TO_PLAYER)
    {
      gPriorFrameEnemyMissileOnscreen = 1;  // used to show condition yellow

      // move spy ship
      dir = gSpriteState[SPY_SPRITE] & SPS_DIRECTION;
      if ((FRAME & 7) == 1) // change direction every 8th frame
      {
        if ((getRandom32() & 0x07) == 0)
        {
          // turn 1/8th the time
          if (getRandom32() & 0x80)
            dir++;
          else
            dir--;
          
          dir &= SPS_DIRECTION;
        }
        gSpriteState[SPY_SPRITE] &= ~SPS_DIRECTION;
        gSpriteState[SPY_SPRITE] |= dir;
        gSpriteImageID[SPY_SPRITE] = SPY_ID + (dir >> 1);
      }
      gSpriteX[SPY_SPRITE] += (gSpySpeed * speed_dir_x[dir]) / 128;
      gSpriteY[SPY_SPRITE] += (gSpySpeed * speed_dir_y[dir]) / 128;
    }
  }
  else
  {
    // work is spread across frames due to performance reasons
    // FRAME & 3 FEATURE
    // --------- ------------
    //     0     spawn I-Type missile
    //     1     spawn P-Type missile
    //     2     spawn Formation
    //     3     spawn Spy Ship
    if (gSpyTimer)
      gSpyTimer--;
    else if (   (FRAME & 3) == 3      // spawn spy ship frame? (see comment above)
             && !zil.gConditionRed    // make sure not already in Condition Red
             && gCurrentSample == 0   // make sure sample not already playing
             && (gSpriteState[PLAYER_SPRITE] & SPS_DEADLY_TO_ENEMY))  // make sure player's alive
    {
      // spawn Spy Ship in front of Star Fighter
      // if StarFighter is pointing diagonally then randomly pick from the 2 sides
      SpawnEnemyShip(SPY_ID, SPY_SPRITE, zil.gShipDir/2 + (zil.gShipDir & getRandom32() & 1));
//      switch((getRandom32() >> 16) & 3)
//      {
//        case 0:
//          x = 159;    // come in on the right side
//          y = (getRandom32() & 128) + 8 ; //(159 * (getRandom32() >> 24)) >> 8;
//          dir = 17;
//          break;
//          
//        case 1:
//          x = 1;     // come in on the left side
//          y = (getRandom32() & 128) + 8 ; // (159 * (getRandom32() >> 24)) >> 8;
//          dir = 1;
//          break;
//          
//        case 2:
//          x = (getRandom32() & 128) + 16 ; // (159 * (getRandom32() >> 24)) >> 8;
//          y = 159;    // come in on the bottom
//          dir = 25;
//          break;
//          
//        case 3:
//          x = (getRandom32() & 128) + 16 ; // (159 * (getRandom32() >> 24)) >> 8;
//          y = -13;     // come in on the top
//          dir = 9;
//          break;
//      }
//      dir += (14 * (getRandom32() >> 24)) >> 8;
//      dir &= SPS_DIRECTION;
//      
//      gSpriteState[SPY_SPRITE]    = SPS_DEADLY_TO_PLAYER | SPS_ENEMYSHIP | dir;
//      gSpriteX[SPY_SPRITE]        = (x << X_FRACTION_BITS) + gWindowX;
//      gSpriteY[SPY_SPRITE]        = (y << Y_FRACTION_BITS) + gWindowY;
//      gSpriteImageID[SPY_SPRITE]  = SPY_ID + (dir >> 1);
//      gSpriteColor[SPY_SPRITE]    = gObjectColors[COLOR_SPY];
      TriggerSample(SAMPLE_SPY_SHIP_SIGHTED);
      gSpyPasses = 3 - MM_LEVEL;  // 4 passes for kids, 3 for easy, 2 for normal, 1 for hard
      gSpyTimer = ResetTimer(gSpyMinTime, gSpyRangeTime);
    }
  }
}

static void NextLevel()
{
  gSector++;
  zil.gInitLevel = 1;
  gRoundStart = 1;
}

const unsigned char pod_off_y[] =
{
  0,  // Horizontal Pod 0
  9,  // Horizontal Pod 1
  40, // Horizontal Pod 2
  53, // Horizontal Pod 3
  39, // Horizontal Pod 4
  9,  // Horizontal Pod 5
  2,  // Vertical Pod 0
  1,  // Vertical Pod 1
  17, // Vertical Pod 2
  39, // Vertical Pod 3
  39, // Vertical Pod 4
  17  // Vertical Pod 5
};


static void PodHit(int station, int pod)
{
  int i;
  int y;
  int podID;
  int height;
  unsigned char * pod_xor;
  
  // pods are numbered 0-5 in this arrangement around the station cores:
  //  0     0 1
  // 5 1   5   2  <- Vertical
  // 4 2    4 3
  //  3  <---------- Horizontal
  
 // if (pod < 0 || pod >= 6)
 //   return;
  
  pod += gSpriteImageID[station] * 6;     // ID = 0 for Horizontal, 1 for Vertical
  
  podID = pod + HORIZONTAL_POD_DIFF_ID;
  height  = gImageHeight[podID];
  pod_xor   = (unsigned char *)(0+gImageGraphics[podID]);
  
  station *= STATION_HEIGHT;
  
  // put xor data into staging area
  y =pod_off_y[pod];
  for (i=0;i<height;i++)
    gStationImages[8 * STATION_HEIGHT + i + y]    |= pod_xor[i];
}

const unsigned char energy_offset[2][2] = // {28, 21} - tops of H and V core
{
  // down   up          station type
  {   32,   32, },  //  horizontal
  {   21,   35  }   //  vertical
};

const unsigned char energy_mask[]   = {0xCF, 0xE7};

static void OpenStation(int station, int amount)
{
  int i;
  int mask;
  int offset1;
  int offset2;
  
  if (amount)
  {
    mask    = energy_mask[gSpriteImageID[station]];
    offset1 = energy_offset[gSpriteImageID[station]][0] + station * STATION_HEIGHT;
    offset2 = energy_offset[gSpriteImageID[station]][1] + station * STATION_HEIGHT;
    
    if (amount < 2)
      gStation[station].state &= ~STS_CORE_OPEN;  // turn off open flag
    else
      gStation[station].state |= STS_CORE_OPEN;   // turn on open flag
    
    for(i=0;i<amount;i++)
    {
      gStationImages[offset1 + i] &= mask;
      gStationImages[offset2 - i] &= mask;
    }
  }
}

static void PrepStationGraphics()
{
 
  int i;
  int j;
  int count;
  
  if (zil.gRoundCleared)
  {
    // zil.gRoundCleared:
    //  0 = at least 1 station still stands
    //  1 = all stations have been destroyed,
    //      waiting for last station explosion sequence to finish
    //  2 = explosion sequence has finished, trigger the tune
    //  3 = wait for end of tune
    
    // make sure player didn't take out last station by colliding with it with their last ship
    if (!(gSpriteState[PLAYER_SPRITE] & SPS_GAMEOVER))
    {
      if (zil.gRoundCleared == 1 && gSpriteState[zil.gStationExplosion] == 0)
        zil.gRoundCleared = 2;  // trigger tune
      
      if (zil.gRoundCleared == 2)
        if (TriggerTune(TUNE_ROUND_CLEAR_A, TUNE_ROUND_CLEAR_B))
          zil.gRoundCleared = 3;  // wait for end of tune
      
      if (zil.gRoundCleared == 3)
        if (SOUND_VOL[gActiveSound[0]] == 0 && SOUND_VOL[gActiveSound[1]] == 0) // check that tune has finished
          NextLevel();
    }

    return;
  }
  
  if (gIdleConditionRedTimer)
  {
    gIdleConditionRedTimer--;
  }
  else if (!zil.gConditionRed)
  {
    // trigger condition red if player been idly flying around the level
    TriggerSample(SAMPLE_CONDITION_RED);
    zil.gConditionRed = 1;
  }
  
  
  count = 0;
  for (i=0;i<8;i++)
  {
    if (gSpriteState[i])
    {
      count++;
      // fresh copy of station image
      myMemcpy((unsigned char *)DD_BASE + STATION_IMAGES + i*STATION_HEIGHT,
               (unsigned char *)0 + gImageGraphics[gSpriteImageID[i]],
               gImageHeight[gSpriteImageID[i]]);

      // manipulate graphics for energy pod
      OpenStation(i, gStation[i].center & SCS_AMOUNT);
      
      if (gStation[i].state & STS_ACTIVE_PODS)
      {
        // init  RAM for xor data
        myMemset((unsigned char *)DD_BASE + STATION_IMAGES + 8*STATION_HEIGHT, 0, STATION_HEIGHT);
        
        // fill xor data for destroyed pods
        for (j=0;j<6;j++)
        {
          if ((gStation[i].state & (1 << j)) == 0)
            PodHit(i, j);
        }
        
        // apply xor to remove destroyed pods
        for (j=0;j<STATION_HEIGHT;j++)
          gStationImages[i * STATION_HEIGHT + j] ^= gStationImages[8 * STATION_HEIGHT + j];
      }
    }
  }
  if (count == 0)
  {
    // remove all enemy ships
    for(i=FORMATION_SPRITE;i<MAX_SPRITES;i++)
      if (gSpriteState[i] && gSpriteImageID[i] >= ETYPE_ID)
        gSpriteState[i] = 0;
    
    // remove all missiles
    for(i=0;i<MAX_MISSILES;i++)
      gMissileState[i] = 0;
    
    zil.gRoundCleared = 1;
  }
}

const unsigned char etype_xoffset[2] = {10,11};
const unsigned char etype_yoffset[2] = {25,21};
const unsigned char etype_imageid[2] = {ETYPE_ID + 4, ETYPE_ID};


static void StationOpenClose()
{
  int s;
  int x;
  int y;
  int dir;
  
//  only process 1 station per frame;
  s = FRAME & 7;

  // open/close at a set rate of speed
  if ((FRAME & gOpenCloseMask) == s)
  {
    if (gSpriteState[s])
    {
      if (!(gStation[s].center & SCS_ALWAYSOPEN))
      {
        if (gStation[s].center & SCS_OPENING) // core is opening
        {
          // if opening = 0 then create E-Type missile if appropriate
          if (((gStation[s].center & SCS_AMOUNT) == 0) && gLaunchEtype)
          {
            gStation[s].state |= STS_CORE_MISSILE;
            // flag as onscreen so move logic won't remove it right away
            gSpriteState[s + ETYPE_SPRITE] = SPS_DEADLY_TO_PLAYER | SPS_SPRITE_ONSCREEN | SPS_ENEMYSHIP;
            gSpriteX[s + ETYPE_SPRITE] = gSpriteX[s] + (etype_xoffset[gSpriteImageID[s]] << X_FRACTION_BITS);
            gSpriteY[s + ETYPE_SPRITE] = gSpriteY[s] + (etype_yoffset[gSpriteImageID[s]] << Y_FRACTION_BITS);
            gSpriteImageID[s + ETYPE_SPRITE] = etype_imageid[gSpriteImageID[s]];
            gSpriteColor[s + ETYPE_SPRITE] = gObjectColors[COLOR_E];
            if (gSpriteImageID[s])
            {
              // vertical station
              y = (gSpriteY[s + ETYPE_SPRITE] >> Y_FRACTION_BITS) - (gSpriteY[PLAYER_SPRITE] >> Y_FRACTION_BITS);
              if (y > (WORLD_HEIGHT / 2)) y -= WORLD_HEIGHT;
              if (y < 0)                  gSpriteImageID[s + ETYPE_SPRITE] += 8;
            }
            else
            {
              // horizontal station
              x = (gSpriteX[s + ETYPE_SPRITE] >> X_FRACTION_BITS) - (gSpriteX[PLAYER_SPRITE] >> X_FRACTION_BITS);
              if (x > (WORLD_WIDTH / 2))  x -= WORLD_WIDTH;
              if (x > 0)                  gSpriteImageID[s + ETYPE_SPRITE] += 8;
            }
          } // end create E-Type missile
          
          // if fully open, launch E-Type and start closing the core
          if ((gStation[s].center & SCS_AMOUNT) == SCS_FULLY_OPEN)
          {
            gStation[s].center &= ~SCS_OPENING;
            gStation[s].state &= ~STS_CORE_MISSILE;
          }
          else
          {
            gStation[s].center++;
          }
        }
        else  // core is closing
        {
          if ((gStation[s].center & SCS_AMOUNT) == SCS_FULLY_CLOSED)
          {
            // only trigger open sequence if the station's E-Type missile is free
            if (gSpriteState[s + ETYPE_SPRITE] == 0)
              gStation[s].center = SCS_OPENING + SCS_FULLY_CLOSED;
          }
          else
          {
            gStation[s].center--;
          }
        }
      }
    }
  }
  
  gPriorFrameEnemyMissileOnscreen = 0;
  
  // but move all launched e-type missiles every frame
  for(s=ETYPE_SPRITE;s<ETYPE_SPRITE+8;s++)
  {
    if (gSpriteState[s])
    {
      if (gSpriteState[s] & SPS_SPRITE_ONSCREEN)
      {
        gPriorFrameEnemyMissileOnscreen = 1;
        if ((gSpriteState[s] & SPS_DEADLY_TO_PLAYER) && !(gStation[s-8].state & STS_CORE_MISSILE))
        {
          dir = (gSpriteImageID[s] - ETYPE_ID) * 2;
          if (dir >= 0 && dir < 32)
          {
            gSpriteX[s] += 2 * speed_dir_x[dir];
            gSpriteY[s] += 2 * speed_dir_y[dir];
          }
        }
      }
      else
      {
        // sprite offscreen, so remove it from play
        gSpriteState[s] = 0;
        // make sure core is visible
        gStation[s-ETYPE_SPRITE].state &= ~STS_CORE_MISSILE;
        
      }
    }
  }
}

const unsigned char double_nybble[] = { 0x00, 0x03, 0x0C, 0x0F,
  0x30, 0x33, 0x3C, 0x3F,
  0xC0, 0xC3, 0xCC, 0xCF,
  0xF0, 0xF3, 0xFC, 0xFF};

static int DoubleBits(int bits)
{
  return (double_nybble[(bits & 0xf0) >> 4] << 8) + double_nybble[bits & 0x0f];
}

static int Collision(int image1_id, int left1, int top1, int sprite_index)
{
  // test if image1_id at position left1, top1 has collided with sprite_index
  // image1_id    = index for graphics & height
  // left1        = top-left corner of sprite, already had >> X_FRACTION_BITS done
  // top1         = top-left corner of sprite, already had >> Y_FRACTION_BITS done
  // sprite_index = index to check against
  //
  //  NOTE: calling routine must make sure sprite_index is for an active sprite
  
  int right1;
  int bottom1;
  
  int left2;
  int top2;
  int right2;
  int bottom2;
  int image2_id;
  int double2;
  int shifting2;
  
  int top3;
  int bottom3;
  
  int j;
  
  unsigned char * image1;
  unsigned char * image2;
  unsigned char * hmove2;
  
  int y_overlap;
  int x_shift;
  int shift;
  
  right1 = left1 + 8;
  
  left2 = gSpriteX[sprite_index] >> X_FRACTION_BITS;
  
  if (left2 < right1)
  {
    image2_id = gSpriteImageID[sprite_index];
    // use double pixel for 2x sprites
    double2 = (gSpriteState[sprite_index] & SPS_2X_SPRITE)  ? 8 : 0;
    shifting2 = (image2_id <= SHIFTING_SPRITES) ? 7 : 0;
    right2 = left2 + 8 + double2 + shifting2;
    
    if (right2 > left1)
    {
      bottom1 = top1 + gImageHeight[image1_id];
      top2 = gSpriteY[sprite_index] >> Y_FRACTION_BITS;
      if (top2 < bottom1)
      {
        bottom2 = top2 + gImageHeight[image2_id];
        if (bottom2 > top1)
        {
          // at this point we've determined that the sprite rectangles overlap
          
          if (top1 > top2)
          {
            top3 = top1;
            image1 = (unsigned char *)0 + gImageGraphics[image1_id];
            image2 = (unsigned char *)0 + gImageGraphics[image2_id] + top1 - top2;
            if (shifting2)
              hmove2 = (unsigned char *)0 + gImageShift[image2_id] + top1 - top2;
          }
          else
          {
            top3 = top2;
            image1 = (unsigned char *)0 + gImageGraphics[image1_id] + top2 - top1;
            image2 = (unsigned char *)0 + gImageGraphics[image2_id];
            if (shifting2)
              hmove2 = (unsigned char *)0 + gImageShift[image2_id];
          }
          
          bottom3 = (bottom1 < bottom2) ? bottom1 : bottom2;
          
          x_shift = left2 - left1;
          y_overlap = bottom3-top3;
          
          if (double2)
          {
            if (x_shift < 0)
            {
              for(j=0;j<y_overlap;j++)
                if ((image1[j] << 8) & (DoubleBits(image2[j]) << -x_shift))
                  return 1;   // collision!
            }
            else
            {
              x_shift += 8;
              for(j=0;j<y_overlap;j++)
                if ((image1[j] << x_shift) & DoubleBits(image2[j]))
                  return 1;   // collision!
            }
          }
          else
          {
            if (x_shift < 0)
            {
              for(j=0;j<y_overlap;j++)
              {
                if (shifting2)
                  shift = hmove2[j] & 7;
                else
                  shift = 0;
                
                if ((image1[j] << shift) & (image2[j] << -x_shift))
                  return 1;   // collision!
              }
            }
            else
            {
              for(j=0;j<y_overlap;j++)
              {
                if (shifting2)
                  shift = hmove2[j] & 7;
                else
                  shift = 0;
                if ((image1[j] << x_shift) & (image2[j] >> shift))
                  return 1;   // collision!
              }
            }
          }
        }
      }
    }
  }
  return 0;
}

const unsigned char quad_pair[] = {0, 0x0f, 0xf0, 0xff};

static unsigned int QuadBits(int bits, int shift)
{
  // must cast as unsigned int so sign isn't maintained if there's a shift
  // ie: if bit 31 is on, we want it off after the shift
  return (unsigned int)((quad_pair[bits & 3]) +
                        (quad_pair[(bits >> 2) & 3] << 8) +
                        (quad_pair[(bits >> 4) & 3] << 16) +
                        (quad_pair[(bits >> 6) & 3] << 24)) >> shift;
}


static int CollisionQuad(int image1_id, int left1, int top1, int sprite_index)
{
  int right1;
  int bottom1;
  
  int left2;
  int top2;
  int right2;
  int bottom2;
  int image2_id;
  
  int top3;
  int bottom3;
  
  int j;
  
  unsigned char * image1;
  unsigned char * image2;
  unsigned char * shift;
  
  int y_overlap;
  int x_shift;
  
  unsigned int prior;
  unsigned int current;
  unsigned int next;
  
  right1 = left1 + 8;
  bottom1 = top1 + gImageHeight[image1_id];
  
  left2 = gSpriteX[sprite_index] >> X_FRACTION_BITS;
  image2_id = gSpriteImageID[sprite_index];
  if (left2 < right1)
  {
    right2 = left2 + 32;
    if (right2 > left1)
    {
      top2 = gSpriteY[sprite_index] >> Y_FRACTION_BITS;
      if (top2 < bottom1)
      {
        bottom2 = top2 + gImageHeight[image2_id];
        if (bottom2 > top1)
        {
          // at this point we've determined that the sprite rectangles overlap
          
          if (top1 > top2)
          {
            top3 = top1;
            image1 = (unsigned char *)0 + gImageGraphics[image1_id];
            image2 = (unsigned char *)(unsigned char *)DD_BASE + STATION_IMAGES + sprite_index*STATION_HEIGHT + top1 - top2;
            shift = (unsigned char *)0 + gImageShift[image2_id] + top1 - top2;
            prior = QuadBits(*(image2-1), *(shift-1) & 7);
          }
          else
          {
            top3 = top2;
            image1 = (unsigned char *)0 + gImageGraphics[image1_id] + top2 - top1;
            image2 = (unsigned char *)DD_BASE + STATION_IMAGES + sprite_index*STATION_HEIGHT;
            shift = (unsigned char *)0 + gImageShift[image2_id];
            prior = 0;  // at top of station
          }
          
          bottom3 = (bottom1 < bottom2) ? bottom1 : bottom2;
          
          x_shift = left2 - left1;
          y_overlap = bottom3-top3;
          
          next = QuadBits(*(image2), *(shift) & 7);
          
          if (x_shift < 0)
          {
            for(j=0;j<y_overlap;j++)
            {
              current = next;
              next = QuadBits(image2[j+1], shift[j+1] & 7);
              //                            if ((image1[j] << 24) & (QuadBits(image2[j], shift[j] & 7) << -x_shift))
              if ((image1[j] << 24) & ((current | (prior & next)) << -x_shift))
                return 1;   // collision!
              prior = current;
            }
          }
          else
          {
            x_shift += 24;
            for(j=0;j<y_overlap;j++)
            {
              current = next;
              next = QuadBits(image2[j+1], shift[j+1] & 7);
              //                            if ((image1[j] << x_shift) & QuadBits(image2[j], shift[j] & 7))
              if ((image1[j] << x_shift) & (current | (prior & next)))
                return 1;   // collision!
              prior = current;
            }
          }
        }
      }
    }
  }
  return 0;
}


static void ShipCollisions()
{
  int i;
  int j;
  int x;
  int y;
  int start;
  int end;
  int sprite;
  
#ifdef RIGHT_DIFFICULTY_FREEZE
  if (SWCHB & 0x80)   // Right Difficulty A = stationary ship
    return;
#endif
  
  if (gSpriteState[PLAYER_SPRITE] & SPS_DEADLY_TO_ENEMY) // make sure player is not exploding
  {
    x = gSpriteX[PLAYER_SPRITE] >> X_FRACTION_BITS;
    y = gSpriteY[PLAYER_SPRITE] >> Y_FRACTION_BITS;
    start = ((y / 32) * STATIONARY_DENSITY) + FIRST_STATIONARY_SPRITE;
    end = (((y + gImageHeight[gSpriteImageID[PLAYER_SPRITE]]) / 32) * STATIONARY_DENSITY) + FIRST_STATIONARY_SPRITE + STATIONARY_DENSITY;
    
    // check for player - station collision
    for (j=0; j<8; j++)
    {
      // station collisions are deadly, station explosions are not
      if (gSpriteState[j] & SPS_SPRITE_ONSCREEN)
      {
        if (CollisionQuad(gSpriteImageID[PLAYER_SPRITE], x, y, j))
        {
          PlayerDestroyed();
          StationDestroyed(j);
        }
      }
    }
    
    // check for player - enemy ship collision
    for(j=ETYPE_SPRITE;j<FIRST_STATIONARY_SPRITE;j++)
    {
      if (j != PLAYER_SPRITE)
      {
        if (gSpriteState[j] & SPS_DEADLY_TO_PLAYER)
        {
          if (Collision(gSpriteImageID[PLAYER_SPRITE], x, y, j))
          {
            PlayerDestroyed();
            if (gSpriteImageID[j] >= ETYPE_ID && gSpriteImageID[j] < SPY_ID+16)
              EnemyDestroyed(j);
          }
        }
      }
    }
    
    // check for player - stationary object collision
    for (j=start; j<end; j++)
    {
      if (j >= MAX_SPRITES)
        sprite = j - (32 * STATIONARY_DENSITY);     // this occurs if shot is straddling the vertical screen wrap point
      else
        sprite = j;
      if (gSpriteState[sprite] & SPS_DEADLY_TO_PLAYER)
      {
        if (Collision(gSpriteImageID[PLAYER_SPRITE], x, y, sprite))
        {
          PlayerDestroyed();
          if (gSpriteImageID[sprite] == MINE_ID)
            MineDestroyed(sprite);
          else if (gSpriteImageID[sprite] >= ASTEROID_ID && gSpriteImageID[sprite] <= ASTEROID_ID+2)
            AsteroidDestroyed(sprite);
        }
      }
    }
    
    // check for Enemy Missile/spy ship and - stationary object collision
    for(i=ETYPE_SPRITE;i<FIRST_STATIONARY_SPRITE;i++)
    {
      if ((i != PLAYER_SPRITE) && (gSpriteState[i] & SPS_ENEMYSHIP) && (gSpriteState[i] & SPS_SPRITE_ONSCREEN))
      {
        x = gSpriteX[i] >> X_FRACTION_BITS;
        y = gSpriteY[i] >> Y_FRACTION_BITS;
        start = ((y / 32) * STATIONARY_DENSITY) + FIRST_STATIONARY_SPRITE;
        end = (((y + gImageHeight[gSpriteImageID[i]]) / 32) * STATIONARY_DENSITY) + FIRST_STATIONARY_SPRITE + STATIONARY_DENSITY;
        for (j=start; j<end; j++)
        {
          if (j >= MAX_SPRITES)
            sprite = j - (32 * STATIONARY_DENSITY);     // this occurs if shot is straddling the vertical screen wrap point
          else
            sprite = j;
          if (gSpriteState[sprite] & SPS_DEADLY_TO_ENEMY)
          {
            if (Collision(gSpriteImageID[i], x, y, sprite))
            {
              EnemyDestroyed(i);
              if (gSpriteImageID[sprite] == MINE_ID)
                MineDestroyed(sprite);
              else if (gSpriteImageID[sprite] >= ASTEROID_ID && gSpriteImageID[sprite] <= ASTEROID_ID+2)
                AsteroidDestroyed(sprite);
            }
          }
        }
      }
    }
  }
}

const signed char shot_dir_x[] = { 0, (7<<6)/12, 1<<6, (7<<6)/12, 0, -(7<<6)/12, -1<<6, -(7<<6)/12 }; // 12:7 pixels

const signed char shot_dir_y[] = {-1<<6,-1<<6, 0, 1<<6, 1<<6, 1<<6, 0,-1<<6 };

static void MoveMissiles()
{
  int i;
  for (i=0;i<MAX_MISSILES;i++)
  {
    
    if (gMissileState[i])
    {
      
#ifdef RIGHT_DIFFICULTY_FREEZE
      if ((SWCHB & 0x80) == 0)
#endif
      {
        if (i < 4)
        {
          // players shots use a 8 direction table.  It moves by a different
          // amount than the 32 direction table so that horizontal and vertical
          // shots take same amount of time to leave the screen
          // (for sound effect cadence)
          gMissileX[i] += 3*shot_dir_x[gMissileDirSpeed[i]] << 1;
          gMissileY[i] += 3*shot_dir_y[gMissileDirSpeed[i]];
        }
        else
        {
          // station shots use 32 direction table
          // station missiles move at 3/4 speed of player's ship
          gMissileX[i] += (speed_dir_x[gMissileDirSpeed[i]]*3)/4;
          gMissileY[i] += (speed_dir_y[gMissileDirSpeed[i]]*3)/4;
          // animate station shots
          gMissileState[i] = 0x80 + SHOT_STATION_ID + ((i + (FRAME >> 2)) & 3);
        }
      }
    }
  }
}

static void MissileCollisions()
{
  int i;
  int j;
  int k;
  int x;
  int y;
  int start;
  int end;
  int sprite;
  int shot_id;
  int pod;
  
#ifdef RIGHT_DIFFICULTY_FREEZE
  if (SWCHB & 0x80)   // Right Difficulty A = stationary ship
    return;
#endif
  
#ifdef DIRECTION_DIAGNOSTICS
  for (i=0;i<FIRST_STATION_MISSILE;i++)
#else
  for (i=0;i<MAX_MISSILES;i++)
#endif
  {
    if (gMissileState[i])
    {
      shot_id = (gMissileState[i] & 0x0f) + FIRST_MISSILE_GRAPHIC_ID;
      x = gMissileX[i] >> X_FRACTION_BITS;
      y = gMissileY[i] >> Y_FRACTION_BITS;
      start = ((y / 32) * STATIONARY_DENSITY) + FIRST_STATIONARY_SPRITE;
      end = (((y + gImageHeight[shot_id]) / 32) * STATIONARY_DENSITY) + FIRST_STATIONARY_SPRITE + STATIONARY_DENSITY;
      
      if (i < 4)
      {
        // only player shots can damage station
        for (j=0; j<8; j++)
        {
          if (gSpriteState[j] & SPS_SPRITE_ONSCREEN)
          {
            pod = gSpriteImageID[j] * 7;  // horizontal = 0, vertical = 7
            
            // test for core hit first if station is open
            if ((gStation[j].state & (STS_CORE_MISSILE | STS_CORE_OPEN)) == STS_CORE_OPEN)
            {
              gSpriteX[MAX_SPRITES]      = gSpriteX[j] + (pod_offset_x[pod+6] << X_FRACTION_BITS);
              gSpriteY[MAX_SPRITES]      = gSpriteY[j] + (pod_offset_y[pod+6] << Y_FRACTION_BITS);
              gSpriteImageID[MAX_SPRITES]= pod_shape[pod+6];
              if (Collision(shot_id, x, y, MAX_SPRITES))
              {
                gIdleConditionRedTimer = gIdleConditionRedMax;  // reset idle timer
                StationDestroyed(j);
                goto missile_hit;
              }
            }
            
            if (CollisionQuad(shot_id, x, y, j))
            {
              for(k=0;k<6;k++)
              {
                if (gStation[j].state & (1 << k)) // if pod is active
                {
                  gSpriteX[MAX_SPRITES]      = gSpriteX[j] + (pod_offset_x[pod+k] << X_FRACTION_BITS);
                  gSpriteY[MAX_SPRITES]      = gSpriteY[j] + (pod_offset_y[pod+k] << Y_FRACTION_BITS);
                  gSpriteImageID[MAX_SPRITES]= pod_shape[pod+k];
                  
                  if (Collision(shot_id, x, y, MAX_SPRITES))
                  {
                    gStation[j].state ^= (1 << k); // turn off pod
                    IncreaseScore(0x200);
                    TriggerSoundEffect(SOUND_CANNON_HIT_ID);
                    gIdleConditionRedTimer = gIdleConditionRedMax;  // reset idle timer
                    // display pod explosion, provided we have a free sprite
                    if (NextUnusedExplosionSprite())
                    {
                      InitDeathSequence(gUnusedExplosionSprite,
                                        EXPLOSION_ASTEROIDPOD_ID,
                                        gSpriteX[MAX_SPRITES],
                                        gSpriteY[MAX_SPRITES],
                                        SAFE_DEBRIS);
                    }
                    // did we destroy the last pod?
                    if ((gStation[j].state & STS_ACTIVE_PODS) == 0)
                      StationDestroyed(j);
                    
                    goto missile_hit;
                  }
                }
              }
              // if we get here the shot dissipates against the station's superstructure
              goto missile_hit;
            }
          }
        }
      }
      else
      {
        // only station shots can damage player's ship
        if (gSpriteState[PLAYER_SPRITE] & SPS_DEADLY_TO_ENEMY)
        {
          // check for station shot - player collision
          if (Collision(SHOT_STATION_ID + FIRST_MISSILE_GRAPHIC_ID,
                        gMissileX[i] >> X_FRACTION_BITS,
                        gMissileY[i] >> Y_FRACTION_BITS,
                        PLAYER_SPRITE))
          {
            PlayerDestroyed();
            goto missile_hit;
          }
        }
      }
      
      
      // check for enemy ship collision
      for(j=ETYPE_SPRITE;j<FIRST_STATIONARY_SPRITE;j++)
      {
        if (j != PLAYER_SPRITE)
        {
          if (gSpriteState[j] & SPS_DEADLY_TO_PLAYER)
          {
            if (Collision(shot_id, x, y, j))
            {
              EnemyDestroyed(j);
              goto missile_hit;
            }
          }
        }
      }
      
      // check for shot - stationary object collision
      for (j=start; j<end; j++)
      {
        if (j >= MAX_SPRITES)
          sprite = j - (32 * STATIONARY_DENSITY);     // this occurs if shot is straddling the vertical screen wrap point
        else
          sprite = j;
        if (gSpriteState[sprite])
        {
          if (Collision(shot_id, x, y, sprite))
          {
            if (gSpriteImageID[sprite] == MINE_ID)
            {
              MineDestroyed(sprite);
              goto missile_hit;
            }
            else if (gSpriteImageID[sprite] >= ASTEROID_ID && gSpriteImageID[sprite] <= ASTEROID_ID+2)
            {
                AsteroidDestroyed(sprite);
                goto missile_hit;
            }
          }
        }
      }
    }
    goto missile_missed;
  missile_hit:
    gMissileState[i] = 0;
  missile_missed:
    ;
  }
}

static void DeathSequences()
{
  int i;
  int j;
  int color;
  for (i=ETYPE_SPRITE; i<MAX_SPRITES; i++)
  {
    if (gSpriteState[i] & SPS_EXPLOSION)
    {
      if (gSpriteState[i] & SPS_EXPLOSION_TIME_MASK)
      {
        if ((FRAME & 1) || (gSpriteState[i] & SPS_FAST_ANIMATION))
        gSpriteState[i]--;
        
        // original explosion time logic
//        gSpriteState[i]--;
//        
//        if (gSpriteState[i] & SPS_FAST_ANIMATION)
//          if (gSpriteState[i] & SPS_EXPLOSION_TIME_MASK)
//            gSpriteState[i]--;
      }
      else
      {
        if (gSpriteState[i] & SPS_FRAME_COUNTDOWN)
        {
          // advance to next frame
          gSpriteState[i] -= SPS_DECREMENT_FRAME;
          gSpriteState[i] |= SPS_EXPLOSION_TIME;
          gSpriteImageID[i]++;
        }
        else
        {
          // animation sequence is complete
          if (gSpriteState[i] & SPS_SHOW_BONUS)
          {
            gSpriteState[i] &= ~SPS_EXPLOSION;
            gSpriteState[i] |= 0x3f;  // show for 64 frames
            if (i == SPY_SPRITE)
              gSpriteImageID[i] = POINTS_200_ID + gSpyBonus;
            else
              gSpriteImageID[i] = POINTS_1500_ID + ((gFormationShipType - ETYPE_ID) / 16);
            gSpriteColor[i] = gObjectColors[COLOR_SHIP];
            gSpriteX[i] -= 2 << X_FRACTION_BITS;
          }
          else
          {
            gSpriteState[i] = 0;
          
            // used to detect when last station explosion is done
            if (zil.gStationExplosion == i)
              zil.gStationExplosion = 0;
          
            if (i == PLAYER_SPRITE)
              PlayerDecreaseLives();
          }
        }
      }
      if (gSpriteState[i] & SPS_EXPLOSION)
      {
        color = (gSpriteImageID[i] - EXPLOSION_PLAYER_ID) * 2 + COLOR_EXPLOSIONS;
        if ((gSpriteState[i] & 8) == 0)
          color++;
        gSpriteColor[i] = gObjectColors[color];
        
        // station explosions are deadly to enemy ships but not the player.
        // They reuse stationary sprites, but aren't in the correct vertical
        // band to be procesed with the normal collision handling.
        if (i >= FIRST_STATIONARY_SPRITE
            && ((gSpriteState[i] & (SPS_DEADLY_TO_PLAYER| SPS_DEADLY_TO_ENEMY)) == SPS_DEADLY_TO_ENEMY))
        {
          for(j=ETYPE_SPRITE;j<FIRST_STATIONARY_SPRITE;j++)
          {
            if (j != PLAYER_SPRITE)
            {
              if (Collision(gSpriteImageID[j],
                            gSpriteX[j] >> X_FRACTION_BITS,
                            gSpriteY[j] >> Y_FRACTION_BITS,
                            i))
                EnemyDestroyed(j);
            }
          }
        }
      }
    }
    else if (gSpriteState[i] & SPS_SHOW_BONUS)
    {
      if (gSpriteState[i] & 0x3f)
        gSpriteState[i]--;
      else
        gSpriteState[i] = 0;
    }
    else if (gSpriteState[i] & SPS_GAMEOVER)
    {
      if (i == PLAYER_SPRITE && gGameDelay)
        gGameDelay--;
      
      if (gGameDelay == 0)
      {
        MODE = 1;   // return to main menu
        gActiveSound[0] = gActiveSound[1] = 0;
        zil.gConditionRed = 0;  // make sure Condition Red doesn't restart
        EndSample();        
        if (zil.gNewHighScore)
          TriggerTune(TUNE_HIGH_SCORE_A, TUNE_HIGH_SCORE_B);
      }
    }
  }
}

static void PlayerDestroyed()
{
  InitDeathSequence(PLAYER_SPRITE,
                    EXPLOSION_PLAYER_ID,
                    gSpriteX[PLAYER_SPRITE] - (4 << X_FRACTION_BITS),
                    gSpriteY[PLAYER_SPRITE] - (8 << Y_FRACTION_BITS),
                    SAFE_DEBRIS);
  TriggerSoundEffect(SOUND_PLAYER_HIT_ID);
  
  
}

static void MineDestroyed(int sprite)
{
  IncreaseScore(0x20);
  InitDeathSequence(sprite,
                    EXPLOSION_MINE_ID,
                    gSpriteX[sprite] - (4 << X_FRACTION_BITS),
                    gSpriteY[sprite] - (8 << Y_FRACTION_BITS),
                    DEADLY_DEBRIS);
  TriggerSoundEffect(SOUND_MINE_ID);
}

static void AsteroidDestroyed(int sprite)
{
  IncreaseScore(0x10);
  InitDeathSequence(sprite,
                    EXPLOSION_ASTEROIDPOD_ID,
                    gSpriteX[sprite],
                    gSpriteY[sprite],
                    SAFE_DEBRIS);
  TriggerSoundEffect(SOUND_ASTEROID_ID);
}

const unsigned short spy_points[] = {0x0200, 0x0400, 0x0800};

static void EnemyDestroyed(int sprite)
{
  int show_bonus;
  int soundeffect;
  int points;
  
  show_bonus = 0;
  soundeffect = 0;
  points = 0;
  
  // START formation kills
  if (sprite == FORMATION_SPRITE)
    zil.gFormationAttackPattern = 0;  // triggers formation to disperse
  
  // station E-Type missile
  if (sprite < PLAYER_SPRITE)
  {
    // is E-Type still docked in the station?
    if (gStation[sprite-8].state & STS_CORE_MISSILE)
    {
      gStation[sprite-8].state &= ~STS_CORE_MISSILE;
      StationDestroyed(sprite-8);
    }
  }
  else if (sprite <SPY_SPRITE)
  {
    soundeffect = SOUND_FORMATION_ID;

    if (--zil.gFormationShipCount == 0)
    {
      // got the bonus points!
      if (gFormationShipType == ETYPE_ID)
        IncreaseScore(0x1500);
      else if (gFormationShipType == PTYPE_ID)
        IncreaseScore(0x1000);
      else  // assume I Type
        IncreaseScore(0x500);
      show_bonus = SPS_SHOW_BONUS;
    }
  }
  // END formation kills
  
  // tally up score and trigger sound effect
  switch ((gSpriteImageID[sprite] - ETYPE_ID) / 16)
  {
    case (ETYPE_ID-ETYPE_ID)/16:
//      if (gSpriteColor[sprite] == gObjectColors[COLOR_E_LEAD])
//        IncreaseScore(0x140);
//      else
//        IncreaseScore(0x70);
//      TriggerSoundEffect(SOUND_ETYPE_HIT_ID);
      points = (gSpriteColor[sprite] == gObjectColors[COLOR_E_LEAD]) ? 0x140 : 0x70;
      if (soundeffect == 0) soundeffect = SOUND_ETYPE_HIT_ID;
      break;
      
    case (ITYPE_ID-ETYPE_ID)/16:
//      if (gSpriteColor[sprite] == gObjectColors[COLOR_I_LEAD])
//        IncreaseScore(0x100);
//      else
//        IncreaseScore(0x50);
//      TriggerSoundEffect(SOUND_ITYPE_SPY_HIT_ID);
      points = (gSpriteColor[sprite] == gObjectColors[COLOR_I_LEAD]) ? 0x100 : 0x50;
      if (soundeffect == 0) soundeffect = SOUND_ITYPE_SPY_HIT_ID;
      
      break;
      
    case (PTYPE_ID-ETYPE_ID)/16:
//      if (gSpriteColor[sprite] == gObjectColors[COLOR_P_LEAD])
//        IncreaseScore(0x120);
//      else
//        IncreaseScore(0x60);
//      TriggerSoundEffect(SOUND_PTYPE_HIT_ID);
      points = (gSpriteColor[sprite] == gObjectColors[COLOR_P_LEAD]) ? 0x120 : 0x60;
      if (soundeffect == 0) soundeffect = SOUND_PTYPE_HIT_ID;
        
      break;
      
    case (SPY_ID-ETYPE_ID)/16:
      show_bonus = SPS_SHOW_BONUS;
      gSpyBonus = (3 * (getRandom32() & 0xff)) >> 8;
//      IncreaseScore(spy_points[gSpyBonus]);
//      TriggerSoundEffect(SOUND_ITYPE_SPY_HIT_ID);
      points = spy_points[gSpyBonus];
      soundeffect = SOUND_ITYPE_SPY_HIT_ID;
      break;
  }
  
  if (points) IncreaseScore(points);
  if (soundeffect)  TriggerSoundEffect(soundeffect);
  
  // trigger explosion sequence
  InitDeathSequence(sprite,
                    EXPLOSION_ENEMYSHIP_ID,
                    gSpriteX[sprite],
                    gSpriteY[sprite],
                    SAFE_DEBRIS);
  
  // flag to show total formation kill or spy ship bonus
  gSpriteState[sprite] |= show_bonus;
}

const unsigned char sd_x_offset[4] =
{
   0, 16,
   0, 16
};

const signed char sd_y_offset[4] =
{
  -2, -2,
  33, 33
};

static void StationDestroyed(int station)
{
  int i;
  
  for (i=0;i<4;i++)
    if (NextUnusedExplosionSprite())
    {
      zil.gStationExplosion = gUnusedExplosionSprite; // used to check if last explosion finished before starting next level
      InitDeathSequence(gUnusedExplosionSprite,
                        EXPLOSION_STATION_ID,
                        gSpriteX[station] + (sd_x_offset[i] << X_FRACTION_BITS),
                        gSpriteY[station] + (sd_y_offset[i] << Y_FRACTION_BITS),
                        SAFE_DEBRIS);
    }
  
  gSpriteState[station] = 0;
  IncreaseScore(0x1500);
  TriggerSoundEffect(SOUND_MINE_ID);
  if (gStation[station].state & STS_CORE_MISSILE)
    EnemyDestroyed(station + ETYPE_SPRITE);
}

static void InitDeathSequence(int sprite, int imageid, unsigned short int x, unsigned short int y, int deadlydebris)
{
  // onscreen flag prevents StationOpenClose() from removing just destroyed E-Type from display
  gSpriteState[sprite] &= SPS_SHOW_BONUS | SPS_SPRITE_ONSCREEN;
  
  if (deadlydebris)
    gSpriteState[sprite] |= SPS_EXPLOSION | SPS_DEADLY_TO_PLAYER | SPS_DEADLY_TO_ENEMY;
  else
    gSpriteState[sprite] |= SPS_EXPLOSION;
  
  if (imageid == EXPLOSION_PLAYER_ID)
  {
    // start player explosion in the "middle" of the first frame (as only 1 color used for it)
    gSpriteState[sprite] |= SPS_4_FRAME_EXPLOSION | (SPS_EXPLOSION_TIME >> 1);
    gSpriteColor[sprite] = gObjectColors[COLOR_EXPLOSIONS + 1];
  }
  else
  {
    if (imageid == EXPLOSION_ASTEROIDPOD_ID)
      gSpriteState[sprite] |= SPS_3_FRAME_EXPLOSION | SPS_EXPLOSION_TIME | SPS_FAST_ANIMATION;
    else if (imageid == EXPLOSION_STATION_ID)
      gSpriteState[sprite] |= SPS_3_FRAME_EXPLOSION | SPS_EXPLOSION_TIME | SPS_2X_SPRITE | SPS_DEADLY_TO_ENEMY;
    else
      gSpriteState[sprite] |= SPS_3_FRAME_EXPLOSION | SPS_EXPLOSION_TIME;
    gSpriteColor[sprite] = gObjectColors[(imageid - EXPLOSION_MINE_ID) * 2 + COLOR_EXPLOSIONS + 8];
  }
  
  gSpriteImageID[sprite] = imageid;
  gSpriteX[sprite] = x;
  gSpriteY[sprite] = y;
}

static int BCDadd(register int a, register int b)
{
  // http://www.cs.uiowa.edu/~jones/bcd/bcd.html
  unsigned int t1, t2, t3, t4, t5, t6;
  
  t1 = a + 0x06666666;
  t2 = t1 + b;
  t3 = t1 ^ b;
  t4 = t2 ^ t3;
  t5 = ~t4 & 0x11111110;
  t6 = (t5 >> 2) | (t5 >> 3);
  return t2 - t6;
}

static void IncreaseScore(int points)
{
  gScore = BCDadd(gScore, points);
  if (gScore >= gExtraLife)
  {
    TriggerSoundEffect(SOUND_BONUS_LIFE);
    gReserveLives++;
    if (gExtraLife == FIRST_BONUS)      gExtraLife = NEXT_BONUS;
    else                                gExtraLife = BCDadd(gExtraLife, NEXT_BONUS);
  }
}

static void PlayerDecreaseLives()
{
  int i;
  int p;
//  int l;
//  const unsigned char *level;
  
  if (gReserveLives == 0)
  {
    gIgnoreFireInMenu = JOY_FIRE_PRESSED;
    gLastScore = gScore;
    if (gScore > gHighScore)
    {
      gHighScore = gScore;
      zil.gNewHighScore = 1;        // flags to play tune
#ifdef SAVE_HIGH_SCORES
      if (MM_INITIAL_SECTOR == 0)
        gHighScores[MM_QUADRANT][MM_LEVEL] = gScore;
#endif
    }

    gSpriteState[PLAYER_SPRITE] = SPS_GAMEOVER;
    gSpriteImageID[PLAYER_SPRITE] = GAMEOVER_ID;
    gSpriteColor[PLAYER_SPRITE] = gObjectColors[COLOR_SHIP];
    gGameDelay = 4*60;
  }
  else
  {
    for (i=ETYPE_SPRITE;i<FIRST_STATIONARY_SPRITE;i++)
      gSpriteState[i] = 0;

    gReserveLives--;
    
//    i = gSector;
//    
//    while (i >= sector_counts[MM_QUADRANT])
//      i -= sector_repeat[MM_QUADRANT];
//    
//    level = levels[MM_QUADRANT][i];
    
//    // position the player at start of level
//    l = 0;
//    gSpriteState[PLAYER_SPRITE]     = SPS_DEADLY_TO_ENEMY;
//    gSpriteImageID[PLAYER_SPRITE]   = SHIP_ID;
//    gSpriteColor[PLAYER_SPRITE]     = gObjectColors[COLOR_SHIP];
//    gSpriteX[PLAYER_SPRITE]         = (level[l++] * 2) << X_FRACTION_BITS;
//    gSpriteY[PLAYER_SPRITE]         = (level[l++] * 4) << Y_FRACTION_BITS;
//    zil.gShipDir                        = 0;
//    zil.gPlayerSpeed                    = 0;
//    
//    gWindowX = gSpriteX[PLAYER_SPRITE] - (80 << X_FRACTION_BITS);
//    gWindowY = gSpriteY[PLAYER_SPRITE] - (((DATASTREAM_SIZE - SHIP_HEIGHT) /2) << Y_FRACTION_BITS);
    
    // set player's new position, but only if they didn't destroy their
    // ship by ramming into the last station on the level.  If they did that
    // then PrepStationGraphics() will trigger the start of the next level.
    if (!zil.gRoundCleared)
    {
      SetPlayerPosition();
      
      for (i=0;i<MAX_MISSILES;i++)
        gMissileState[i] = 0;
      
      for (i=0;i<8;i++)
        for (p=0;p<6;p++)
          gStation[i].podcooldown[p] = 0xffff;
      
      zil.gFormationAttackPattern = 0;
      zil.gFormationShipCount = 0;
      //    zil.gConditionRed = 0; now in SetPlayerPosition()
    }
  }
}


static void NextUnusedMissile(int *missile)
{
  while (gMissileState[*missile] && *missile < MAX_MISSILES)
    (*missile)++;
  if (*missile == MAX_MISSILES)
    *missile = 0; // 0 means no unused missile
}

const unsigned char pod_neighbor[7] = {5, 0, 1, 2, 3, 4, 5};

static void StationMissiles()
{
  int s;
  int p;
  int x;
  int y;
  int i;
  int distance[6];
  int min_pod[2];
  int id;
  int next_missile;
  int player_x;
  int player_y;
  
  p = 0;
  for(s=0;s<8;s++)
  {
    p |= gSpriteState[s] & SPS_SPRITE_ONSCREEN;
  }
  
  if (gNotSoChatty)
    gNotSoChatty--;
  
  if (!gNotSoChatty                       // make sure Alert! Alert! is not said too frequently
      && p                                // make sure a station is onscreen
      && gPriorFrameStationOnscreen != p  // and none were onscreen before
      && (gSpriteState[PLAYER_SPRITE] & SPS_DEADLY_TO_ENEMY)) // make sure player's alive
  {
    gAlertRepeat = 1;
    TriggerSample(SAMPLE_ALERT_ALERT);
    gNotSoChatty = 60 * 10; // say at most once every 10 seconds
  }
  gPriorFrameStationOnscreen = p;
  
  // target the middle of the player
  player_x = ((gSpriteX[PLAYER_SPRITE] >> X_FRACTION_BITS) + 3);
  player_y = ((gSpriteY[PLAYER_SPRITE] >> Y_FRACTION_BITS) + 7);
  
#ifndef DIRECTION_DIAGNOSTICS
  
  next_missile = FIRST_STATION_MISSILE;
  NextUnusedMissile(&next_missile);
  
  // For performance reasons, shoot from at most 2 of the stations per frame.
  // If need be, change to this for 1 station per frame:
  //    S=FRAME & 7
  for(s=FRAME & 3;s<8;s+=4)
  {
    if (gSpriteState[s])  // if station's still active
    {
      // cool off the guns
      for(p=0;p<6;p++)
        if (gStation[s].podcooldown[p] > gPodCoolDownRate)
          gStation[s].podcooldown[p] -= gPodCoolDownRate;
        else
          gStation[s].podcooldown[p] = 0;
      
      // if station's on screen and we have a free missile
      if ((gSpriteState[s] & SPS_SPRITE_ONSCREEN) && next_missile)
      {
        id = gSpriteImageID[s] * 7;
        min_pod[0] = 0;
        
        for(p=0;p<6;p++)
        {
          // test using position shot is fired from, not the pod location in pod_offset_?
          x = (gSpriteX[s] >> X_FRACTION_BITS) + pod_shot_offset_x[id + p];
          y = (gSpriteY[s] >> Y_FRACTION_BITS) + pod_shot_offset_y[id + p];
          x -= player_x;
          y -= player_y;
          
          distance[p] = (x * x) + (y * y);
          if (distance[p] < distance[min_pod[0]])
            min_pod[0] = p;
        }
        
        p = min_pod[0];
        min_pod[1] = distance[pod_neighbor[min_pod[0]]] < distance[pod_neighbor[min_pod[0]+2]] ? pod_neighbor[min_pod[0]] : pod_neighbor[min_pod[0]+2];
        
        for(i=0; i<2 && next_missile;i++)
        {
          p = min_pod[i];
          if (gStation[s].podcooldown[p] == 0 && (gStation[s].state & (1 << p)))
            if ((getRandom32()>>24) < gPodFireChance)
            {
              gMissileState[next_missile]     = 0x80 + SHOT_STATION_ID + ((next_missile + (FRAME >> 2)) & 3);
              gMissileX[next_missile]         = gSpriteX[s] + (pod_shot_offset_x[id + p] << X_FRACTION_BITS);
              gMissileY[next_missile]         = gSpriteY[s] + (pod_shot_offset_y[id + p] << Y_FRACTION_BITS);
              gMissileDirSpeed[next_missile]  = DirectionToPlayer(gMissileX[next_missile],
                                                                  gMissileY[next_missile],
                                                                  pod_shot_init_dir[id + p],
                                                                  DTP_MISSILE,
                                                                  DTP_DIRECTION);
              gStation[s].podcooldown[p] = 0xffff;
              NextUnusedMissile(&next_missile);
            }
        }
      }
    }
  }
#else
  int dir;
  
  // diagnostic mode
  // find pod closest to player for onscreen stations
  for(s=0;s<8;s++)
  {
    
    gSpriteState[ETYPE_SPRITE + s] = 0;
    gMissileState[FIRST_STATION_MISSILE + s] = 0;
    gMissileState[FIRST_STATION_MISSILE + s + 8] = 0;
    gMissileState[FIRST_STATION_MISSILE + s + 16] = 0;
    gMissileState[FIRST_STATION_MISSILE + s + 24] = 0;
    gMissileState[FIRST_STATION_MISSILE + s + 32] = 0;
    
    if (gSpriteState[s] & SPS_SPRITE_ONSCREEN)
    {
      id = gSpriteImageID[s] * 7;
      min_pod[0] = 0;
      
      for(p=0;p<6;p++)
      {
        // test using position shot is fired from, not the pod location in pod_offset_?
        x = (gSpriteX[s] >> X_FRACTION_BITS) + pod_shot_offset_x[id + p];
        y = (gSpriteY[s] >> Y_FRACTION_BITS) + pod_shot_offset_y[id + p];
        x -= player_x;
        y -= player_y;
        
        distance[p] = (x * x) + (y * y);
        if (distance[p] < distance[min_pod[0]])
          min_pod[0] = p;
      }
      
      p = min_pod[0];
      min_pod[1] = distance[pod_neighbor[min_pod[0]]] < distance[pod_neighbor[min_pod[0]+2]] ? pod_neighbor[min_pod[0]] : pod_neighbor[min_pod[0]+2];

      // show closest pod for diagnostics
      gSpriteState[ETYPE_SPRITE + s]    = SPS_DEADLY_TO_PLAYER;
      gSpriteX[ETYPE_SPRITE + s]        = gSpriteX[s] + (pod_offset_x[id + min_pod[0]] << X_FRACTION_BITS);
      gSpriteY[ETYPE_SPRITE + s]        = gSpriteY[s] + (pod_offset_y[id + min_pod[0]] << Y_FRACTION_BITS);
      gSpriteColor[ETYPE_SPRITE + s]    = 0x0d;     // white
      gSpriteImageID[ETYPE_SPRITE + s]  = POD_ID;
      
      // show 2nd closest pod for diagnostics
      gSpriteState[MISC_SPRITES + s]    = SPS_DEADLY_TO_PLAYER;
      gSpriteX[MISC_SPRITES + s]        = gSpriteX[s] + (pod_offset_x[id + min_pod[1]] << X_FRACTION_BITS);
      gSpriteY[MISC_SPRITES + s]        = gSpriteY[s] + (pod_offset_y[id + min_pod[1]] << Y_FRACTION_BITS);
      gSpriteColor[MISC_SPRITES + s]    = 0x0d;     // white
      gSpriteImageID[MISC_SPRITES + s]  = POD_ID;
      
      // station shots use 32 direction table
      gMissileX[FIRST_STATION_MISSILE + s] = gSpriteX[s] + (pod_shot_offset_x[id + min_pod[0]] << X_FRACTION_BITS);
      gMissileY[FIRST_STATION_MISSILE + s] = gSpriteY[s] + (pod_shot_offset_y[id + min_pod[0]] << Y_FRACTION_BITS);
      // animate station shots
      gMissileState[FIRST_STATION_MISSILE + s] = 0x80 + SHOT_STATION_ID + ((s + (FRAME >> 2)) & 3);
      
      dir = DirectionToPlayer(gMissileX[FIRST_STATION_MISSILE + s],
                              gMissileY[FIRST_STATION_MISSILE + s],
                              pod_shot_init_dir[id + min_pod[0]],
                              DTP_MISSILE,
                              DTP_DIRECTION);
      
      // station shots use 32 direction table
      gMissileX[FIRST_STATION_MISSILE + 8 + s] = gMissileX[FIRST_STATION_MISSILE + s] + 8*speed_dir_x[dir];
      gMissileY[FIRST_STATION_MISSILE + 8 + s] = gMissileY[FIRST_STATION_MISSILE + s] + 8*speed_dir_y[dir];
      gMissileState[FIRST_STATION_MISSILE + 8 + s] = 0x80 + SHOT_STATION_ID + ((s + (FRAME >> 2)) & 3);
      
      // station shots use 32 direction table
      gMissileX[FIRST_STATION_MISSILE +16 + s] = gMissileX[FIRST_STATION_MISSILE + s] + 16*speed_dir_x[dir];
      gMissileY[FIRST_STATION_MISSILE +16 + s] = gMissileY[FIRST_STATION_MISSILE + s] + 16*speed_dir_y[dir];
      gMissileState[FIRST_STATION_MISSILE +16 + s] = 0x80 + SHOT_STATION_ID + ((s + (FRAME >> 2)) & 3);
      
      // station shots use 32 direction table
      gMissileX[FIRST_STATION_MISSILE +24 + s] = gMissileX[FIRST_STATION_MISSILE + s] + 32*speed_dir_x[dir];
      gMissileY[FIRST_STATION_MISSILE +24 + s] = gMissileY[FIRST_STATION_MISSILE + s] + 32*speed_dir_y[dir];
      gMissileState[FIRST_STATION_MISSILE +24 + s] = 0x80 + SHOT_STATION_ID + ((s + (FRAME >> 2)) & 3);
      
      // station shots use 32 direction table
      gMissileX[FIRST_STATION_MISSILE +32 + s] = gMissileX[FIRST_STATION_MISSILE + s] + 64*speed_dir_x[dir];
      gMissileY[FIRST_STATION_MISSILE +32 + s] = gMissileY[FIRST_STATION_MISSILE + s] + 64*speed_dir_y[dir];
      gMissileState[FIRST_STATION_MISSILE +32 + s] = 0x80 + SHOT_STATION_ID + ((s + (FRAME >> 2)) & 3);
    }
  }
#endif
  
  
  
//  int i;
//  int f;
//  int s;
//  int p;
//  int count;
//  
//  s = -1;
//  for (i=0;i<8;i++)
//  {
//    if (gSpriteState[i] & SPS_STATION_COOLDOWN)               // cooldown active
//      gSpriteState[i]--;
//    else if ((gSpriteState[i] & (SPS_SPRITE_ONSCREEN | SPS_STATION_COOLDOWN)) == SPS_SPRITE_ONSCREEN)    // if onscreen & cooldown then we can shoot
//      s = i;
//  }
//  
//  f = 0;
//  for (i=4;i<MAX_MISSILES;i++)
//    if (gMissileState[i] == 0)
//      f = i;
//  
//  if (f && s != -1)
//  {
//    gSpriteState[s] |= SPS_STATION_COOLDOWN;
//    gMissileState[f] = 0x80 + SHOT_STATION_ID + (f&3);
//    gMissileDirSpeed[f] = getRandom32() & 7;
//    count = 0;
//    for (i=0;i<6;i++)
//      if (gStation[i].state & (1 << i))
//        count++;    // count active pods
//    count = (count * (getRandom32() & 0xff)) >> 8;          // random 0 - (count-1)
//    
//    p = gSpriteImageID[s] * 7;
//    for (i=0;i<6;i++)
//      if (gStation[i].state & (1 << i))
//      {
//        count--;
//        if (!count)
//          p += i;
//      }
//    gMissileX[f] = gSpriteX[s] + ((pod_offset_x[p]+3) << X_FRACTION_BITS);
//    gMissileY[f] = gSpriteY[s] + ((pod_offset_y[p]+7) << Y_FRACTION_BITS);
//  }
}

static int distance(int x, int y, int object, int dir) {
  int player_x = gSpriteX[PLAYER_SPRITE];
  int player_y = gSpriteY[PLAYER_SPRITE];
  
  int test_x = x + speed_dir_x[dir] - player_x;
  int test_y = y + speed_dir_y[dir] - player_y;
  
  if (object == DTP_MISSILE)
  {
    // missile targets center of player
    player_x += 3;
    player_y += 7;
  }
  
  // handle wrap around the edges of the screen
  if (test_x < 0) test_x = -test_x;
  if (test_y < 0) test_y = -test_y;
  if (test_x > ((WORLD_WIDTH << X_FRACTION_BITS)/ 2)) test_x = (WORLD_WIDTH << X_FRACTION_BITS) - test_x;
  if (test_y > ((WORLD_HEIGHT<< Y_FRACTION_BITS)/ 2)) test_y = (WORLD_HEIGHT << Y_FRACTION_BITS) - test_y;
  
  return (test_x * test_x) + (test_y * test_y);
}

static int TurnAwayFromPlayer(int x, int y, /*int object,*/ int dir)
{
//  int player_x;
//  int player_y;
  int i;
  int max_distance;
  int max_dir;
  int test_dir;
//  int test_x;
//  int test_y;
  int test_distance;
  
  max_distance = 0;
  max_dir = dir;
  
//  player_x = gSpriteX[PLAYER_SPRITE];
//  player_y = gSpriteY[PLAYER_SPRITE];
  
  for(i=dir-1; i<=dir+1;i++)
  {
    test_dir = i & 31;
    
//    test_x = x + speed_dir_x[test_dir] - player_x;
//    test_y = y + speed_dir_y[test_dir] - player_y;
//    // handle wrap around the edges of the screen
//    
//    if (test_x < 0) test_x = -test_x;
//    if (test_y < 0) test_y = -test_y;
//    if (test_x > ((WORLD_WIDTH << X_FRACTION_BITS)/ 2)) test_x = (WORLD_WIDTH << X_FRACTION_BITS) - test_x;
//    if (test_y > ((WORLD_HEIGHT<< Y_FRACTION_BITS)/ 2)) test_y = (WORLD_HEIGHT << Y_FRACTION_BITS) - test_y;
//    
//    test_distance = (test_x * test_x) + (test_y * test_y);
    
    test_distance = distance(x, y, DTP_SPRITE, test_dir);
    if (test_distance > max_distance)
    {
      max_distance = test_distance;
      max_dir = test_dir;
    }
  }
  return max_dir;
}

static int DirectionToPlayer(int x, int y, int dir, int object, int turn_mode)
{
  // find direction towards player from (x,y).  dir is current direction, passed
  // so as to limit the number of checks that need to be made
  // object is DTP_SPRITE or DTP_MISSILE which adjusts for the smaller missile
  
//  int player_x;
//  int player_y;
  int min_distance;
//  int test_x;
//  int test_y;
  int test_dir;
  int test_distance;
  int rotation;
  int done;
  int first;
  
//  player_x = gSpriteX[PLAYER_SPRITE];
//  player_y = gSpriteY[PLAYER_SPRITE];
//  
//  if (object == DTP_MISSILE)
//  {
//    // missile targets center of player
//    player_x += 3;
//    player_y += 7;
//  }
//
//  test_x = x + speed_dir_x[dir] - player_x;
//  test_y = y + speed_dir_y[dir] - player_y;
//  
//  // handle wrap around the edges of the screen
//  if (test_x < 0) test_x = -test_x;
//  if (test_y < 0) test_y = -test_y;
//  if (test_x > ((WORLD_WIDTH << X_FRACTION_BITS)/ 2)) test_x = (WORLD_WIDTH << X_FRACTION_BITS) - test_x;
//  if (test_y > ((WORLD_HEIGHT<< Y_FRACTION_BITS)/ 2)) test_y = (WORLD_HEIGHT << Y_FRACTION_BITS) - test_y;
//  
//  min_distance = (test_x * test_x) + (test_y * test_y);

  min_distance = distance(x, y, object, dir);
  
  done = 0;
  rotation = 1;
  first = 1;
  do
  {
    test_dir = (dir + rotation) & 31;
    
//    test_x = x + speed_dir_x[test_dir] - player_x;
//    test_y = y + speed_dir_y[test_dir] - player_y;
//    
//    // handle wrap around the edges of the screen
//    if (test_x < 0) test_x = -test_x;
//    if (test_y < 0) test_y = -test_y;
//    if (test_x > ((WORLD_WIDTH << X_FRACTION_BITS)/ 2)) test_x = (WORLD_WIDTH << X_FRACTION_BITS) - test_x;
//    if (test_y > ((WORLD_HEIGHT<< Y_FRACTION_BITS)/ 2)) test_y = (WORLD_HEIGHT << Y_FRACTION_BITS) - test_y;
//    
//    test_distance = (test_x * test_x) + (test_y * test_y);
    
    test_distance = distance(x, y, object, test_dir);
    
    if (test_distance < min_distance)
    {
      min_distance = test_distance;
      dir = test_dir;
      if (turn_mode) return dir;
    }
    else if (first)
    {
      rotation = -1;
    }
    else
    {
      done = 1;
    }
      
    first = 0;
  } while (!done);
  
  return dir;
}



static void MoveEnemyShips()
{
  int i;
  int type;
  int count[2];
//  int x;
//  int y;
  int dir;
  
//  // wait for player to be at full speed
//  if (zil.gPlayerSpeed < 128 || zil.gRoundCleared || gRoundStart)
//    return;
  
  // used for determining when to spawn new ships
  count[0] = count[1] = 0;
  
  // move enemy ships
  for(i=MISC_SPRITES;i<FIRST_STATIONARY_SPRITE;i++)
  {
    if (gSpriteState[i] & SPS_ENEMYSHIP)
    {
      if (gSpriteState[i] & SPS_SPRITE_ONSCREEN)
      {
        gPriorFrameEnemyMissileOnscreen = 1;

        // update ship count
        type = (gSpriteImageID[i] - ITYPE_ID) / 16;
        count[type]++;
        
        dir = gSpriteState[i] & SPS_DIRECTION;
        
        if (((i+FRAME) & 7) == 0) // update direction every 8th frame
        {
          if (type && !zil.gConditionRed)
          {
            // p-type ignores player if not in condition red mode
            if ((getRandom32() & 0x07) == 0)
            {
              // turn 1/8th the time
              if (getRandom32() & 0x80)
                dir++;
              else
                dir--;
              
              dir &= SPS_DIRECTION;
            }
          }
          else
          {
            // i-type follows player
            dir = DirectionToPlayer(gSpriteX[i],
                                    gSpriteY[i],
                                    dir,
                                    DTP_SPRITE,
                                    DTP_TURN);
          }
          
          gSpriteState[i] &= ~SPS_DIRECTION;
          gSpriteState[i] |= dir;
          gSpriteImageID[i] = ITYPE_ID + (type * 16) + (dir >> 1);
        }
        
        gSpriteX[i] += (gEnemySpeed * speed_dir_x[dir]) / 128;
        gSpriteY[i] += (gEnemySpeed * speed_dir_y[dir]) / 128;
      }
      else
      {
        gSpriteState[i] = 0;
      }
    }
  }
  
  // spawn new enemy ships
  type = FRAME & 1; // 0 = I-Type, 1 = P-Type
  
  if (gEnemyTimer[type])
    gEnemyTimer[type]--;
  // NOTE: cannot do the following as an ELSE IF due to the gConditionRed always
  // launching new missiles if count allows for it
  
  // FRAME & 3 FEATURE
  // --------- ------------
  //     0     spawn I-Type missile
  //     1     spawn P-Type missile
  //     2     spawn Formation
  //     3     spawn Spy Ship
  if (   (FRAME & 3) == type                        // spawn X-Type frame? (see comment above)
      && (!gEnemyTimer[type] || zil.gConditionRed)  // timer done OR condition red ?
      && (count[type] < gEnemyMax[type]))           // not hit the max?
  {
    if (NextUnusedEnemySprite())
    {
      SpawnEnemyShip(ITYPE_ID + (type * 16), gUnusedEnemySprite, (getRandom32() >> 16) & 3);
      gEnemyTimer[type] = ResetTimer(gEnemyMinTime, gEnemyRangeTime);
    }
  }
}



static void InitUnusedSprites()
{
  gUnusedExplosionSprite  = MAX_SPRITES - 1;
  gUnusedEnemySprite      = MISC_SPRITES;
}

static int NextUnusedExplosionSprite()
{
  while (gSpriteState[gUnusedExplosionSprite] && gUnusedExplosionSprite > MISC_SPRITES)
    gUnusedExplosionSprite--;
  return (gSpriteState[gUnusedExplosionSprite] == 0);
}

static int NextUnusedEnemySprite()
{
  // for performance, enemy ships are only drawn with sprites MISC_SPRITES thru FIRST_STATIONARY_SPRITE-1
  while (gSpriteState[gUnusedEnemySprite] && gUnusedEnemySprite < FIRST_STATIONARY_SPRITE-1)
    gUnusedEnemySprite++;
  return (gSpriteState[gUnusedEnemySprite] == 0);
}


static unsigned short int ResetTimer(unsigned short int min, unsigned short int range)
{
  return min + ((range * (getRandom32() & 0xffff)) >> 16);
}

static void SpawnEnemyShip(int enemyID, int index, int side)
{
  int x;
  int y;
  int enter;
  
  int dir;
  // side 0=top, 1=right, 2=bottom, 3=left
  
  enter = (getRandom32() & 127) + ((side & 1) ? 8 : 16);
  switch(side)
  {
    case 0:       // come in on the top
      x = enter;
      y = -13;
      dir = 9;
      break;
      
    case 1:       // come in on the right side
      x = 159;
      y = enter;
      dir = 17;
      break;
      
    case 2:       // come in on the bottom
      x = enter;
      y = 159;
      dir = 25;
      break;
      
    default:      // come in on the left side
      x = 1;
      y = enter;
      dir = 1;
      break;
  }
  gSpriteX[index]        = (x << X_FRACTION_BITS) + gWindowX;
  gSpriteY[index]        = (y << Y_FRACTION_BITS) + gWindowY;
  
  if (enemyID == SPY_ID)
  {
    dir = DirectionToPlayer(gSpriteX[index],
                            gSpriteY[index],
                            dir,
                            DTP_SPRITE,
                            DTP_DIRECTION);
  }
  else
  {
    dir += (14 * (getRandom32() >> 24)) >> 8;
    dir &= SPS_DIRECTION;
  }
  
  gSpriteState[index]    = SPS_DEADLY_TO_PLAYER | SPS_ENEMYSHIP | dir;
  gSpriteImageID[index]  = enemyID + (dir >> 1);
  gSpriteColor[index]    = gObjectColors[COLOR_I + ((enemyID-ITYPE_ID)/16)];
}
